"""Plots a violin plot of the evaluation metrics."""

import pandas as pd
import plotly.graph_objects as go

from ..pipeline_function import pipeline_function
from .vioheat import VioHeat


class Violin(VioHeat):
    """Plots a violin plot of the evaluation metrics."""

    @pipeline_function
    def __init__(self, df: pd.DataFrame, normalized_metrics=False, whole_name=False):
        """
        Initialize the Violin class.

        :param df: either containing all predictions for all algorithms or all tests for one algorithm (including
            robustness, randomization, … tests then)
        :param normalized_metrics: whether the metrics are normalized
        :param whole_name: whether the whole name should be displayed
        """
        super().__init__(df, normalized_metrics, whole_name)
        self.df["box"] = self.df["algorithm"] + "_" + self.df["rand_setting"] + "_" + self.df["test_mode"]
        # remove columns with only NaN values
        self.df = self.df.dropna(axis=1, how="all")
        self.fig = go.Figure()
        self.occurring_metrics = [metric for metric in self.all_metrics if metric in self.df.columns]

    @pipeline_function
    def draw_and_save(self, out_prefix: str, out_suffix: str) -> None:
        """
        Draw the violin and save it to a file.

        :param out_prefix: e.g., results/my_run/violin_plots/
        :param out_suffix: e.g., algorithms_normalized
        """
        self._draw()
        path_out = f"{out_prefix}violin_{out_suffix}.html"
        self.fig.write_html(path_out)

    def _draw(self) -> None:
        self._create_evaluation_violins()
        count_sum = (
            self.count_r2
            + self.count_pearson
            + self.count_spearman
            + self.count_kendall
            + self.count_mse
            + self.count_rmse
            + self.count_mae
        )
        buttons_update = list(
            [
                dict(
                    label="All Metrics",
                    method="update",
                    args=[
                        {"visible": [True] * count_sum},
                        {"title": "All Metrics"},
                    ],
                ),
                dict(
                    label="R^2",
                    method="update",
                    args=[
                        {"visible": [True] * self.count_r2 + [False] * (count_sum - self.count_r2)},
                        {"title": "R^2"},
                    ],
                ),
                dict(
                    label="Pearson",
                    method="update",
                    args=[
                        {
                            "visible": [False] * self.count_r2
                            + [True] * self.count_pearson
                            + [False] * (count_sum - self.count_r2 - self.count_pearson)
                        },
                        {"title": "Pearson"},
                    ],
                ),
                dict(
                    label="Spearman",
                    method="update",
                    args=[
                        {
                            "visible": [False] * (self.count_r2 + self.count_pearson)
                            + [True] * self.count_spearman
                            + [False] * (count_sum - self.count_r2 - self.count_pearson - self.count_spearman)
                        },
                        {"title": "Spearman"},
                    ],
                ),
                dict(
                    label="Kendall",
                    method="update",
                    args=[
                        {
                            "visible": [False] * (self.count_r2 + self.count_pearson + self.count_spearman)
                            + [True] * self.count_kendall
                            + [False]
                            * (
                                count_sum
                                - self.count_r2
                                - self.count_pearson
                                - self.count_spearman
                                - self.count_kendall
                            )
                        },
                        {"title": "Kendall"},
                    ],
                ),
            ]
        )
        if not self.normalized_metrics:
            buttons_update += list(
                [
                    dict(
                        label="MSE",
                        method="update",
                        args=[
                            {
                                "visible": [False] * (count_sum - self.count_mse - self.count_rmse - self.count_mae)
                                + [True] * self.count_mse
                                + [False] * (self.count_rmse + self.count_mae)
                            },
                            {"title": "MSE"},
                        ],
                    ),
                    dict(
                        label="RMSE",
                        method="update",
                        args=[
                            {
                                "visible": [False] * (count_sum - self.count_rmse - self.count_mae)
                                + [True] * self.count_rmse
                                + [False] * self.count_mae
                            },
                            {"title": "RMSE"},
                        ],
                    ),
                    dict(
                        label="MAE",
                        method="update",
                        args=[
                            {"visible": [False] * (count_sum - self.count_mae) + [True] * self.count_mae},
                            {"title": "MAE"},
                        ],
                    ),
                ]
            )
        self.fig.update_layout(
            updatemenus=[
                dict(
                    active=0,
                    buttons=buttons_update,
                )
            ]
        )
        self.fig.update_layout(title_text="All Metrics", height=600, width=1100)

    def _create_evaluation_violins(self):
        print("Drawing Violin plots ...")
        self.count_r2 = 0
        self.count_pearson = 0
        self.count_spearman = 0
        self.count_kendall = 0
        self.count_mse = 0
        self.count_rmse = 0
        self.count_mae = 0
        for metric in self.occurring_metrics:
            if "R^2" in metric:
                self.count_r2 += 1 * len(self.df["box"].unique())
            elif "Pearson" in metric:
                self.count_pearson += 1 * len(self.df["box"].unique())
            elif "Spearman" in metric:
                self.count_spearman += 1 * len(self.df["box"].unique())
            elif "Kendall" in metric:
                self.count_kendall += 1 * len(self.df["box"].unique())
            elif "RMSE" in metric:
                self.count_rmse += 1 * len(self.df["box"].unique())
            elif "MSE" in metric:
                self.count_mse += 1 * len(self.df["box"].unique())
            elif "MAE" in metric:
                self.count_mae += 1 * len(self.df["box"].unique())
            self._add_violin(metric)

    def _add_violin(self, metric):
        for box in self.df["box"].unique():
            tmp_df = self.df[self.df["box"] == box]
            if self.whole_name:
                label = box + ": " + metric
            else:
                label = box.split("_")[0] + ": " + metric
            self.fig.add_trace(
                go.Violin(
                    y=tmp_df[metric],
                    x=[label] * len(tmp_df[metric]),
                    name=label,
                    box_visible=True,
                    meanline_visible=True,
                )
            )



"""
Contains the DRPModel class.

The DRPModel class is an abstract wrapper class for drug response prediction models.


"""

import inspect
import os
from abc import ABC, abstractmethod
from typing import Any

import numpy as np
import yaml
from sklearn.model_selection import ParameterGrid

from ..datasets.dataset import DrugResponseDataset, FeatureDataset
from ..pipeline_function import pipeline_function


class DRPModel(ABC):
    """
    Abstract wrapper class for drug response prediction models.

    The DRPModel class is an abstract wrapper class for drug response prediction models.
    It has a boolean attribute is_single_drug_model indicating whether it is a single drug model and a boolean
    attribute early_stopping indicating whether early stopping is used.
    """

    # Used in the pipeline!
    early_stopping = False
    # Then, the model is trained per drug
    is_single_drug_model = False

    @classmethod
    @abstractmethod
    @pipeline_function
    def get_model_name(cls) -> str:
        """
        Returns the name of the model.

        :return: model name
        """

    @classmethod
    @pipeline_function
    def get_hyperparameter_set(cls) -> list[dict[str, Any]]:
        """
        Loads the hyperparameters from a yaml file which is located in the same directory as the model.

        :returns: list of hyperparameter sets
        :raises ValueError: if the hyperparameters are not in the correct format
        :raises KeyError: if the model is not found in the hyperparameters file
        """
        hyperparameter_file = os.path.join(os.path.dirname(inspect.getfile(cls)), "hyperparameters.yaml")

        with open(hyperparameter_file, encoding="utf-8") as f:
            try:
                hpams = yaml.safe_load(f)[cls.get_model_name()]
            except yaml.YAMLError as exc:
                raise ValueError(f"Error in hyperparameters.yaml: {exc}") from exc
            except KeyError as key_exc:
                raise KeyError(f"Model {cls.get_model_name()} not found in hyperparameters.yaml") from key_exc

        if hpams is None:
            return [{}]
        # each param should be a list
        for hp in hpams:
            if not isinstance(hpams[hp], list):
                hpams[hp] = [hpams[hp]]
        grid = list(ParameterGrid(hpams))
        return grid

    @property
    @abstractmethod
    def cell_line_views(self) -> list[str]:
        """
        Returns the sources the model needs as input for describing the cell line.

        :return: cell line views, e.g., ["methylation", "gene_expression", "mirna_expression",
            "mutation"]. If the model does not use cell line features, return an empty list.
        """

    @property
    @abstractmethod
    def drug_views(self) -> list[str]:
        """
        Returns the sources the model needs as input for describing the drug.

        :return: drug views, e.g., ["descriptors", "fingerprints", "targets"]. If the model does not use drug features,
            return an empty list.
        """

    @abstractmethod
    def build_model(self, hyperparameters: dict[str, Any]) -> None:
        """
        Builds the model, for models that use hyperparameters.

        :param hyperparameters: hyperparameters for the model

        Example::

            self.model = ElasticNet(alpha=hyperparameters["alpha"], l1_ratio=hyperparameters["l1_ratio"])
        """

    @abstractmethod
    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "checkpoints",
    ) -> None:
        """
        Trains the model.

        :param output: training data associated with the response output
        :param cell_line_input: input associated with the cell line, required for all models
        :param drug_input: input associated with the drug, optional because single drug models do not use drug features
        :param output_earlystopping: optional early stopping dataset
        :param model_checkpoint_dir: directory to save the model checkpoints
        """

    @abstractmethod
    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the response for the given input.

        :param drug_ids: list of drug ids, also used for single drug models, there it is just an array containing the
            same drug id
        :param cell_line_ids: list of cell line ids
        :param cell_line_input: input associated with the cell line, required for all models
        :param drug_input: input associated with the drug, optional because single drug models do not use drug features
        :returns: predicted response
        """

    @abstractmethod
    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Load the cell line features before the train/predict method is called.

        Required to implement for all models. Could, e.g., call get_multiomics_feature_dataset() or
        load_and_select_gene_features() from models/utils.py.

        :param data_path: path to the data, e.g., data/
        :param dataset_name: name of the dataset, e.g., "GDSC2"
        :returns: FeatureDataset with the cell line features
        """

    @abstractmethod
    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset | None:
        """
        Load the drug features before the train/predict method is called.

        Required to implement for all models that use drug features. Could, e.g.,
        call load_drug_fingerprint_features() or load_drug_ids_from_csv() from models/utils.py.

        For single drug models, this method can return None.

        :param data_path: path to the data, e.g., data/
        :param dataset_name: name of the dataset, e.g., "GDSC2"
        :returns: FeatureDataset or None
        """

    def get_concatenated_features(
        self,
        cell_line_view: str | None,
        drug_view: str | None,
        cell_line_ids_output: np.ndarray,
        drug_ids_output: np.ndarray,
        cell_line_input: FeatureDataset | None,
        drug_input: FeatureDataset | None,
    ) -> np.ndarray:
        """
        Concatenates the features to an input matrix X for the given cell line and drug views.

        :param cell_line_view: gene expression, methylation, etc.
        :param drug_view: ids, fingerprints, etc.
        :param cell_line_ids_output: cell line ids
        :param drug_ids_output: drug ids
        :param cell_line_input: input associated with the cell line
        :param drug_input: input associated with the drug
        :returns: X, the feature matrix needed for, e.g., sklearn models
        :raises ValueError: if no features are provided

        This can, e.g., be done in the training method to produce a large input feature matrix for the model where
        the rows are the samples and the columns are the cell line and drug features concatenated. This method is an
        alternative to using DataLoaders. It is used for models operating on the whole input matrix at once.

        Example::

            x = self.get_concatenated_features(
                cell_line_view="gene_expression",
                drug_view="fingerprints",
                cell_line_ids_output=output.cell_line_ids,
                drug_ids_output=output.drug_ids,
                cell_line_input=cell_line_input,
                drug_input=drug_input,
            )
            self.model.fit(x, output.response)
        """
        inputs = self.get_feature_matrices(
            cell_line_ids=cell_line_ids_output,
            drug_ids=drug_ids_output,
            cell_line_input=cell_line_input,
            drug_input=drug_input,
        )
        if drug_view is not None:
            if drug_view not in inputs:
                raise ValueError(f"Expected drug_view '{drug_view}' to be in inputs, but it was not. Inputs: {inputs}")
        if cell_line_view is not None:
            if cell_line_view not in inputs:
                raise ValueError(
                    f"Expected cell_line_view '{cell_line_view}' to be in inputs, but it was not. Inputs: {inputs}"
                )

        cell_line_features = None if cell_line_view is None else inputs.get(cell_line_view)
        drug_features = None if drug_view is None else inputs.get(drug_view)

        if cell_line_features is not None and drug_features is not None:
            x = np.concatenate((cell_line_features, drug_features), axis=1)
        elif cell_line_features is not None:
            x = cell_line_features
        elif drug_features is not None:
            x = drug_features
        else:
            raise ValueError("No features provided.")
        return x

    def get_feature_matrices(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset | None,
        drug_input: FeatureDataset | None,
    ) -> dict[str, np.ndarray]:
        """
        Returns the feature matrices for the given cell line and drug ids by retrieving the correct views.

        :param cell_line_ids: cell line identifiers
        :param drug_ids: drug identifiers
        :param cell_line_input: cell line omics features
        :param drug_input: drug omics features
        :returns: dictionary with the feature matrices
        :raises ValueError: if the input does not contain the correct views

        This can e.g., done to produce the input for the predict() method for deep learning models:
        Example::

            input_data = self.get_feature_matrices(
                cell_line_ids=cell_line_ids,
                drug_ids=drug_ids,
                cell_line_input=cell_line_input,
                drug_input=drug_input,
            )
            (
                gene_expression,
                mutations,
                cnvs
            ) = (
                input_data["gene_expression"],
                input_data["mutations"],
                input_data["copy_number_variation_gistic"]
            )
            return self.model.predict(gene_expression, mutations, cnvs)

        Or to produce separate inputs for the train()/predict() method for other models if the model does not operate
        on the concatenated input matrix::

            inputs = self.get_feature_matrices(
                cell_line_ids=output.cell_line_ids,
                drug_ids=output.drug_ids,
                cell_line_input=cell_line_input,
                 drug_input=drug_input,
            )
            (
                gene_expression,
                methylation,
                mutations,
                copy_number_variation_gistic,
                fingerprints,
            ) = (
                inputs["gene_expression"],
                inputs["methylation"],
                inputs["mutations"],
                inputs["copy_number_variation_gistic"],
                inputs["fingerprints"],
            )
            self.model.fit(
                gene_expression, methylation, mutations, copy_number_variation_gistic, fingerprints, output.response
            )
        """
        cell_line_feature_matrices = {}
        if cell_line_input is not None:
            for cell_line_view in self.cell_line_views:
                if cell_line_view not in cell_line_input.view_names:
                    raise ValueError(f"Cell line input does not contain view {cell_line_view}")
                cell_line_feature_matrices[cell_line_view] = cell_line_input.get_feature_matrix(
                    view=cell_line_view, identifiers=cell_line_ids
                )
        drug_feature_matrices = {}
        if drug_input is not None:
            for drug_view in self.drug_views:
                if drug_view not in drug_input.view_names:
                    raise ValueError(f"Drug input does not contain view {drug_view}")
                drug_feature_matrices[drug_view] = drug_input.get_feature_matrix(view=drug_view, identifiers=drug_ids)

        return {**cell_line_feature_matrices, **drug_feature_matrices}



"""Includes custom torch.nn.Modules for the DIPK model: AttentionLayer, DenseLayer, Predictor."""

import torch
import torch.nn as nn

from .attention_utils import MultiHeadAttentionLayer

features_dim_gene = 512
features_dim_bionic = 512
DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class AttentionLayer(nn.Module):
    """Custom attention layer for the DIPK model."""

    def __init__(self, heads: int = 1):
        """
        Initialize the attention layer with a multi-head attention layer with a specified number of heads.

        :param heads: number of heads for the multi-head attention layer
        """
        super().__init__()
        self.fc_layer_0 = nn.Linear(features_dim_gene, 768)
        self.fc_layer_1 = nn.Linear(features_dim_bionic, 768)
        self.attention_0 = MultiHeadAttentionLayer(hid_dim=768, n_heads=heads, dropout=0.3, device=DEVICE)
        self.attention_1 = MultiHeadAttentionLayer(hid_dim=768, n_heads=heads, dropout=0.3, device=DEVICE)

    def forward(
        self, molgnet_features: torch.Tensor, mask: torch.Tensor, gene_expression: torch.Tensor, bionic: torch.Tensor
    ) -> torch.Tensor:
        """
        Forward pass of the attention layer.

        :param molgnet_features: MolGNet features
        :param mask: mask for the MolGNet features, as molecules have varying sizes (valid atom features are True)
        :param gene_expression: gene expression features of the graph data
        :param bionic: bionic network features of the graph data
        :returns: tensor of MolGNet features after attention layer
        """
        gene_expression = nn.functional.relu(self.fc_layer_0(gene_expression))  # Shape: [batch_size, feature_dim_gene]
        bionic = nn.functional.relu(self.fc_layer_1(bionic))  # Shape: [batch_size, feature_dim_bionic]

        # Preparing query, key, value for attention layers
        query_0 = torch.unsqueeze(gene_expression, 1)  # Shape: [batch_size, 1, 768] for gene
        query_1 = torch.unsqueeze(bionic, 1)  # Shape: [batch_size, 1, 768] for bionic
        key = molgnet_features  # Shape: [batch_size, seq_len, 768] (features from MolGNet)
        value = molgnet_features  # Shape: [batch_size, seq_len, 768] (same as key)

        mask = torch.unsqueeze(mask, 1).unsqueeze(2)

        # Apply the first attention layer
        x_att = self.attention_0(query_0, key, value, mask)  # Output: [batch_size, seq_len, hid_dim]
        x = torch.squeeze(x_att[0])  # Squeeze to remove the extra dimension (1)

        # Apply the second attention layer
        x_att = self.attention_1(query_1, key, value, mask)  # Output: [batch_size, seq_len, hid_dim]
        x += torch.squeeze(x_att[0])  # Add the result of the second attention to the first

        return x


class DenseLayers(nn.Module):
    """Custom dense layers for the DIPK model."""

    def __init__(self, fc_layer_num: int, fc_layer_dim: list[int], dropout_rate: float):
        """
        Initialize the dense layers of the DIPK model which follow the attention layer.

        :param fc_layer_num: number of fully connected layers
        :param fc_layer_dim: list of dimensions for each fully connected layer
        :param dropout_rate: dropout rate for all fully connected layers
        """
        super().__init__()
        self.fc_layer_num = fc_layer_num
        self.fc_layer_0 = nn.Linear(features_dim_gene, 512)
        self.fc_layer_1 = nn.Linear(features_dim_bionic, 512)
        self.fc_input = nn.Linear(768 + 512, 768 + 512)
        self.fc_layers = torch.nn.Sequential(
            nn.Linear(768 + 512, 512),
            nn.Linear(512, fc_layer_dim[0]),
            nn.Linear(fc_layer_dim[0], fc_layer_dim[1]),
            nn.Linear(fc_layer_dim[1], fc_layer_dim[2]),
            nn.Linear(fc_layer_dim[2], fc_layer_dim[3]),
            nn.Linear(fc_layer_dim[3], fc_layer_dim[4]),
            nn.Linear(fc_layer_dim[4], fc_layer_dim[5]),
        )
        self.dropout_layers = torch.nn.ModuleList([nn.Dropout(p=dropout_rate) for _ in range(fc_layer_num)])
        self.fc_output = nn.Linear(fc_layer_dim[fc_layer_num - 2], 1)

    def forward(self, x: torch.Tensor, gene: torch.Tensor, bionic: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the dense layers.

        :param x: output tensor from the attention layer
        :param gene: gene expression features (GEF) of the graph data
        :param bionic: biological network features (BNF) of the graph data
        :returns: output tensor after the dense layers
        """
        if len(x.shape) == 1:  # happens if the batch size is 1
            x = torch.unsqueeze(x, 0)

        gene = torch.nn.functional.relu(self.fc_layer_0(gene))
        bionic = torch.nn.functional.relu(self.fc_layer_1(bionic))
        f = torch.cat((x, gene + bionic), 1)
        f = torch.nn.functional.relu(self.fc_input(f))
        for layer_index in range(self.fc_layer_num):
            f = torch.nn.functional.relu(self.fc_layers[layer_index](f))
            f = self.dropout_layers[layer_index](f)
        f = self.fc_output(f)
        return f


class Predictor(nn.Module):
    """Whole DIPK model."""

    def __init__(self, heads: int, fc_layer_num: int, fc_layer_dim: list[int], dropout_rate: float):
        """
        Initialize the DIPK model with the specified hyperparameters.

        :param heads: number of heads for the multi-head attention layer
        :param fc_layer_num: number of fully connected layers for the dense layers
        :param fc_layer_dim: number of neurons for each fully connected layer
        :param dropout_rate: dropout rate for all fully connected layers
        """
        super().__init__()
        self.attention_layer = AttentionLayer(heads=heads)
        self.dense_layers = DenseLayers(fc_layer_num=fc_layer_num, fc_layer_dim=fc_layer_dim, dropout_rate=dropout_rate)

    def forward(
        self,
        molgnet_drug_features: torch.Tensor,
        gene_expression: torch.Tensor,
        bionic: torch.Tensor,
        molgnet_mask: torch.Tensor,
    ) -> torch.Tensor:
        """
        Forward pass of the DIPK model.

        :param molgnet_drug_features: tensor of MolGNet features from graph data
        :param gene_expression: gene expression features (GEF) of the graph data
        :param bionic: biological network features (BNF) of the graph data
        :param molgnet_mask: mask for the MolGNet features, as molecules have varying sizes
        :returns: output tensor of the DIPK model
        """
        molgnet_drug_features = self.attention_layer(molgnet_drug_features, molgnet_mask, gene_expression, bionic)
        f = self.dense_layers(molgnet_drug_features, gene_expression, bionic)
        return f



"""Contains the SimpleNeuralNetwork model."""

import platform
import warnings

import numpy as np
from sklearn.preprocessing import StandardScaler

from drevalpy.datasets.dataset import DrugResponseDataset, FeatureDataset

from ..drp_model import DRPModel
from ..utils import load_and_select_gene_features, load_drug_fingerprint_features
from .utils import FeedForwardNetwork


class SimpleNeuralNetwork(DRPModel):
    """Simple Feedforward Neural Network model with dropout using only gene expression data."""

    cell_line_views = ["gene_expression"]
    drug_views = ["fingerprints"]
    early_stopping = True

    def __init__(self):
        """Initializes the SimpleNeuralNetwork.

        The model is build in train(). The gene_expression_scalar is set to the StandardScaler() and later fitted
        using the training data only.
        """
        super().__init__()
        self.model = None
        self.hyperparameters = None
        self.gene_expression_scaler = StandardScaler()

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: SimpleNeuralNetwork
        """
        return "SimpleNeuralNetwork"

    def build_model(self, hyperparameters: dict):
        """
        Builds the model from hyperparameters.

        :param hyperparameters: includes units_per_layer and dropout_prob.
        """
        self.hyperparameters = hyperparameters

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "checkpoints",
    ) -> None:
        """
        First scales the gene expression data and trains the model.

        The gene expression data is first arcsinh transformed. Afterward, the StandardScaler() is fitted on the
        training gene expression data only. Then, it transforms all gene expression data.
        :param output: training data associated with the response output
        :param cell_line_input: cell line omics features
        :param drug_input: drug omics features
        :param output_earlystopping: optional early stopping dataset
        :param model_checkpoint_dir: directory to save the model checkpoints
        :raises ValueError: if drug_input (fingerprints) is missing

        """
        if drug_input is None:
            raise ValueError("drug_input (fingerprints) are required for SimpleNeuralNetwork.")

        # Apply arcsinh transformation and scaling to gene expression features
        if "gene_expression" in self.cell_line_views:
            cell_line_input.apply(function=np.arcsinh, view="gene_expression")
            self.gene_expression_scaler = cell_line_input.fit_transform_features(
                train_ids=np.unique(output.cell_line_ids),
                transformer=self.gene_expression_scaler,
                view="gene_expression",
            )

        dim_gex = next(iter(cell_line_input.features.values()))["gene_expression"].shape[0]
        dim_fingerprint = next(iter(drug_input.features.values()))["fingerprints"].shape[0]

        self.model = FeedForwardNetwork(
            hyperparameters=self.hyperparameters,
            input_dim=dim_gex + dim_fingerprint,
        )

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=".*does not have many workers which may be a bottleneck.*",
            )
            warnings.filterwarnings(
                "ignore",
                message="Starting from v1\\.9\\.0, `tensorboardX` has been removed.*",
            )
            self.model.fit(
                output_train=output,
                cell_line_input=cell_line_input,
                drug_input=drug_input,
                cell_line_views=self.cell_line_views,
                drug_views=self.drug_views,
                output_earlystopping=output_earlystopping,
                trainer_params={
                    "max_epochs": self.hyperparameters.get("max_epochs", 100),
                    "progress_bar_refresh_rate": 500,
                },
                batch_size=16,
                patience=5,
                num_workers=1 if platform.system() == "Windows" else 8,
                model_checkpoint_dir=model_checkpoint_dir,
            )

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the response for the given input.

        :param cell_line_ids: IDs of the cell lines to be predicted
        :param drug_ids: IDs of the drugs to be predicted
        :param cell_line_input: gene expression of the test data
        :param drug_input: fingerprints of the test data
        :returns: the predicted drug responses
        """
        x = self.get_concatenated_features(
            cell_line_view="gene_expression",
            drug_view="fingerprints",
            cell_line_ids_output=cell_line_ids,
            drug_ids_output=drug_ids,
            cell_line_input=cell_line_input,
            drug_input=drug_input,
        )

        return self.model.predict(x)

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features.

        :param data_path: Path to the gene expression and landmark genes
        :param dataset_name: name of the dataset
        :return: FeatureDataset containing the cell line gene expression features, filtered through the landmark genes
        """
        return load_and_select_gene_features(
            feature_type="gene_expression",
            gene_list="landmark_genes",
            data_path=data_path,
            dataset_name=dataset_name,
        )

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the fingerprint data.

        :param data_path: Path to the fingerprints, e.g., data/
        :param dataset_name: name of the dataset, e.g., GDSC1
        :returns: FeatureDataset containing the fingerprints
        """
        return load_drug_fingerprint_features(data_path, dataset_name, fill_na=True)



"""
Command-line tool to generate and update harmonized tissue annotations for cancer cell lines.

Uses Cellosaurus and DepMap metadata (and some errors in these datasets are fixed manually)
and adds it to the response datasets e.g. for LTO splits.
Use it for reference or to create the tissue annotations for custom datasets

This tool performs the following steps:
1. Loads all unique Cellosaurus IDs from a specified set of drug response datasets (e.g., CCLE, GDSC1).
2. Downloads and parses the latest Cellosaurus reference file to extract metadata
(name, site of derivation, and disease).
3. Loads DepMap sample information to merge and normalize disease labels.
4. Applies a curated tissue synonym dictionary to map specific
diseases to broader, biologically meaningful tissue categories.
5. Manually overrides mappings for key misclassified or ambiguous cell lines
based on verified external sources (e.g., ATCC, NCI).
6. Saves the final tissue mapping to a central CSV file.
7. Propagates the harmonized tissue column back into each of the original datasets.

Arguments:
- `data_path` (str): Path to the directory containing all datasets and metadata files.
- `dataset` (str): One of {"CCLE", "GDSC1", "GDSC2", "CTRPv1", "CTRPv2", "all"}
- save_tissue_mapping (bool): If True, saves the tissue mapping to a CSV file.
or a custom dataset name. When "all" is specified, the tissue mapping is applied across all non-custom datasets

Notes:
- This script assumes each dataset is stored as a CSV file under `<data_path>/<dataset>/<dataset>.csv`.
- Tissue mapping is derived from a curated synonym dictionary and adjusted
for special cases based on literature or database references.
- The output `tissue_mapping.csv` is saved in `<data_path>/meta/`.

Example usage:
    python -m mymodule.add_tissue_mapping data/ all
"""

import argparse
import os
import urllib.request

import pandas as pd

from . import AVAILABLE_DATASETS
from .loader import download_dataset

_tissue_synonyms = {
    "Lung": [
        "lung",
        "small cell lung cancer",
        "lung adenocarcinoma",
        "minimally invasive lung adenocarcinoma",
        "lung squamous cell carcinoma",
        "lung carcinoid tumor",
        "lung mucoepidermoid carcinoma",
        "lung giant cell carcinoma",
        "lung adenosquamous carcinoma",
        "mesothelioma",
        "pleural mesothelioma",
    ],
    "Colon": [
        "colorectal",
        "colon adenocarcinoma",
        "cecum adenocarcinoma",
        "gardner syndrome",
        "intestine",
        "rectal adenocarcinoma",
    ],
    "Small Intestine": [
        "small_intestine",
        "small intestine adenocarcinoma",
        "small intestine carcinoid tumor",
        "small intestine neuroendocrine tumor",
        "small intestine neuroendocrine carcinoma",
    ],
    "Stomach": [
        "gastric",
        "gastric adenocarcinoma",
        "gastric tubular adenocarcinoma",
        "gastric signet ring cell adenocarcinoma",
        "gastric choriocarcinoma",
    ],
    "Esophagus": ["esophagus", "squamous cell carcinoma of the esophagus", "adenocarcinoma of the esophagus"],
    "Head and neck": [
        "upper_aerodigestive",
        "head and neck squamous cell carcinoma",
        "squamous cell carcinoma of the larynx",
        "squamous cell carcinoma of the hypopharynx",
        "squamous cell carcinoma of the oral cavity",
        "squamous cell carcinoma of the oral tongue",
        "oral epithelial dysplasia",
        "parotid gland mucoepidermoid carcinoma",
    ],
    "Skin": [
        "skin",
        "melanoma",
        "cutaneous melanoma",
        "amelanotic melanoma",
        "skin squamous cell carcinoma",
        "vulvar melanoma",
    ],
    "Blood": [
        "leukemia",
        "acute myeloid leukemia",
        "chronic myeloid leukemia",
        "precursor b-cell acute lymphoblastic leukemia",
        "precursor t-cell acute lymphoblastic leukemia",
        "classic hairy cell leukemia",
        "mixed phenotype acute leukemia",
        "acute erythroid leukemia",
        "acute myelomonocytic leukemia",
        "acute monoblastic/monocytic leukemia",
        "hereditary spherocytosis",
        "multiple myeloma",
        "multiple_myeloma",
        "bone marrow",
        "natural killer cell lymphoblastic leukemia/lymphoma",
        "b-lymphoblastic leukemia/lymphoma with t(v",
        "b-lymphoblastic leukemia/lymphoma with t(17",
    ],
    "Lymph": [
        "lymphoma",
        "hodgkin lymphoma",
        "diffuse large b-cell lymphoma",
        "burkitt lymphoma",
        "primary mediastinal large b-cell lymphoma",
        "b-cell non-hodgkin lymphoma",
        "sezary syndrome",
        "follicular lymphoma",
        "alk-positive anaplastic large cell lymphoma",
        "primary effusion lymphoma",
        "splenic marginal zone lymphoma",
        "primary cutaneous t-cell lymphoma",
    ],
    "Bone": ["bone", "osteosarcoma", "ewing sarcoma", "chondrosarcoma"],
    "Muscle": ["rhabdomyosarcoma", "rhabdoid", "embryonal rhabdomyosarcoma", "rhabdoid tumor"],
    "Soft Tissue": [
        "soft_tissue",
        "fibrosarcoma",
        "undifferentiated pleomorphic sarcoma",
        "liposarcoma",
        "fibroblast",
        "synovial sarcoma",
    ],
    "Thyroid": ["thyroid", "anaplastic thyroid carcinoma", "differentiated thyroid carcinoma"],
    "Brain": [
        "central_nervous_system",
        "glioblastoma",
        "gliosarcoma",
        "astrocytoma",
        "anaplastic astrocytoma",
        "diffuse astrocytoma",
    ],
    "Nervous system": [
        "neuroblastoma",
        "peripheral primitive neuroectodermal tumor",
        "peripheral_nervous_system",
        "nervous system",
    ],
    "Breast": [
        "breast",
        "breast carcinoma",
        "breast ductal carcinoma",
        "invasive breast carcinoma of no special type",
    ],
    "Ovary": [
        "ovary",
        "high grade ovarian serous adenocarcinoma",
        "ovarian serous adenocarcinoma",
        "maligant granulosa cell tumor of the ovary",
    ],
    "Uterus": ["uterus", "high-grade neuroendocrine carcinoma of the cervix uteri"],
    "Cervix": [
        "cervix",
        "vagina",
        "vulvar carcinoma",
        "vulvar squamous cell carcinoma",
        "squamous cell carcinoma of the cervix uteri",
    ],
    "Prostate": ["prostate"],
    "Kidney": ["kidney", "renal cell carcinoma", "clear cell renal carcinoma"],
    "Bladder": ["urinary_tract", "bladder carcinoma"],
    "Liver": [
        "liver",
        "cholangiocarcinoma",
        "bile_duct",
        "hepatoblastoma",
        "carcinoma of gallbladder and extrahepatic biliary tract",
    ],
    "Pancreas": ["pancreas", "pancreatic ductal adenocarcinoma"],
    "Adrenal Gland": ["adrenal_cortex"],
    "Embryonic": ["embryo", "non-central nervous system-localized embryonal carcinoma", "embryonal carcinoma"],
    "Unknown": ["unknown", "other", "carcinoid syndrome"],
}


def _parse_cellosaurus(cellosaurus_path: str) -> tuple[dict, dict, dict]:
    """
    Parse Cellosaurus file and return mappings from cellosaurus ID to name, site, and disease.

    :param cellosaurus_path: Path to the Cellosaurus text file
    :return: Tuple of dictionaries (id_to_name, id_to_site, id_to_disease)
    """
    id_to_name, id_to_site, id_to_disease = {}, {}, {}

    with open(cellosaurus_path, encoding="utf-8") as f:
        current_ids, current_name, site, disease = [], None, None, None
        for line in f:
            if line.startswith("ID   "):
                current_name = line.strip().split("   ")[1]
            elif line.startswith("AC   "):
                current_ids = [s.strip() for s in line[5:].split(";") if s.strip()]
            elif line.startswith("CC   Derived from site:"):
                parts = line.strip().split(":", 1)[1].split(";")
                if len(parts) >= 2:
                    site = parts[1].strip()
            elif line.startswith("DI   ") and current_ids:
                parts = line[5:].split(";")
                if len(parts) >= 3:
                    disease = parts[2].strip()
            elif line.strip() == "//":
                for cid in current_ids:
                    if current_name:
                        id_to_name[cid] = current_name
                    if site:
                        id_to_site[cid] = site
                    if disease:
                        id_to_disease[cid] = disease
                current_ids, current_name, site, disease = [], None, None, None

    return id_to_name, id_to_site, id_to_disease


def _apply_manual_cell_line_corrections(tissue_map: pd.Series) -> pd.Series:
    """Apply manual tissue corrections for misclassified or ambiguous cell lines with documented sources.

    :param tissue_map: Series mapping Cellosaurus IDs to tissues
    :return: Updated tissue mapping Series
    """
    manual_entries = [
        # CVCL_0977: Hs 888.Lu
        # Source: Cell Model Passports - SIDM01745
        # https://cellmodelpassports.sanger.ac.uk/passports/SIDM01745
        ("CVCL_0977", "Lung"),
        # CVCL_1072: ARH-77
        # Source: Culture Collections - ARH-77
        # https://www.culturecollections.org.uk/products/celllines/detail.jsp?refId=88121201
        ("CVCL_1072", "Blood"),
        # CVCL_1305: IM-9
        # Source: ATCC - CCL-159
        # https://www.atcc.org/products/ccl-159
        ("CVCL_1305", "Blood"),
        # CVCL_1665: RPMI-6666
        # Source: https://scicrunch.org/resolver/RRID:CVCL_1665?q=&i=rrid:cvcl_1665-kclb-10113
        # Derived from leukemic cells of a myeloma patient, EBV-transformed B lymphoblastoid line
        ("CVCL_1665", "Blood"),
        # CVCL_0807: Hs 578Bst
        # Source: https://scicrunch.org/resolver/CVCL_0807/
        ("CVCL_0807", "Breast"),
        # CVCL_L296: H-STS
        # Source: https://www.nature.com/articles/s41588-019-0490-z
        ("CVCL_L296", "Blood"),
        # CVCL_ZA06: WT2-iPS
        # Source: https://discover.nci.nih.gov/rsconnect/cellminercdb/cell_lines/wt2ips_cellminercdb.html
        # NOTE: Excluded on purpose — cl doesn't exist, was wrong mapping before
        # ("CVCL_ZA06", "Skin"),
        # CVCL_L298: P-STS
        # Source: Pfragner R, Behmel A, Höger H, Beham A,
        # Ingolic E, Stelzer I, Svejda B, Moser VA, Obenauf AC, Siegl V, et al. (2009).
        # Establishment and characterization of three novel cell lines
        # – P-STS, L-STS, H-STS – derived from a human metastatic midgut carcinoid.
        # Anticancer Research, 29(6), 1951–1961.
        ("CVCL_L298", "Small Intestine"),
        # CVCL_3386 was misclassified.
        # source: https://www.cellosaurus.org/CVCL_3386
        ("CVCL_3386", "Blood"),
    ]

    for cellosaurus_id, tissue in manual_entries:
        tissue_map[cellosaurus_id] = tissue

    return tissue_map


def _harmonize_disease_annotations(df_cellosaurus: pd.DataFrame, sample_info: pd.DataFrame) -> pd.DataFrame:
    """Merge Cellosaurus and DepMap data, normalize names, and harmonize disease annotations.

    :param df_cellosaurus: DataFrame containing Cellosaurus data
    :param sample_info: DataFrame containing DepMap sample information
    :return: Merged DataFrame with harmonized disease annotations

    """
    df_cellosaurus["name_norm"] = df_cellosaurus["cell_line_name"].str.lower().str.replace(r"[^a-z0-9]", "", regex=True)
    sample_info["name_norm"] = (
        sample_info["stripped_cell_line_name"].str.lower().str.replace(r"[^a-z0-9]", "", regex=True)
    )

    merged = pd.merge(df_cellosaurus, sample_info, on="name_norm", how="left")
    merged["disease"] = merged["disease"].replace(r"^\s*$", "unknown", regex=True)

    merged["disease_combined"] = (
        merged.apply(
            lambda row: (
                row["cellosaurus_disease"]
                if (pd.isna(row["disease"]) or row["disease"].strip().lower() == "unknown")
                and pd.notna(row["cellosaurus_disease"])
                else row["disease"]
            ),
            axis=1,
        )
        .str.strip()
        .str.lower()
    )

    return merged


def main():
    """Main function to add tissue mapping to datasets."""
    parser = argparse.ArgumentParser(description="Add tissue mapping to datasets")
    parser.add_argument("data_path", help="Path to dataset root directory", default="data")
    parser.add_argument("dataset", help="Dataset name (e.g., CCLE) or 'all'", default="all")
    parser.add_argument(
        "--save_tissue_mapping",
        action="store_true",
        help="Save the tissue mapping to a CSV file",
    )
    args = parser.parse_args()
    data_path = args.data_path
    dataset = args.dataset
    save_tissue_mapping = args.save_tissue_mapping
    if dataset != "all":
        datasets = [dataset]
    else:
        datasets = AVAILABLE_DATASETS.keys()

    cell_lines = []

    # Load all unique Cellosaurus IDs from available datasets
    for ds in datasets:
        csv_path = os.path.join(data_path, ds, f"{ds}.csv")
        try:
            df = pd.read_csv(csv_path, dtype=str, low_memory=False)
            cell_lines.extend(df["cellosaurus_id"].dropna().unique())
        except FileNotFoundError:
            continue

    cellosaurus_ids = pd.Series(cell_lines).drop_duplicates().reset_index(drop=True)

    #  Download Cellosaurus if needed
    cellosaurus_path = os.path.join(data_path, "meta", "cellosaurus.txt")
    cellosaurus_path.parent.mkdir(parents=True, exist_ok=True)

    if not cellosaurus_path.exists():
        url = "https://ftp.expasy.org/databases/cellosaurus/cellosaurus.txt"
        urllib.request.urlretrieve(url, cellosaurus_path)  # noqa-S310

    # Parse Cellosaurus
    id_to_name, id_to_site, id_to_disease = _parse_cellosaurus(cellosaurus_path)

    # Build Cellosaurus DataFrame
    df_cellosaurus = pd.DataFrame(
        {
            "cellosaurus_id": cellosaurus_ids,
            "cell_line_name": cellosaurus_ids.map(id_to_name),
            "cellosaurus_derived_from_site": cellosaurus_ids.map(id_to_site),
            "cellosaurus_disease": cellosaurus_ids.map(id_to_disease),
        }
    ).dropna(subset=["cell_line_name"])

    # Load DepMap sample_info
    depmap_path = os.path.join(data_path, "meta", "DepMap_sample_info.csv")
    if not os.path.exists(depmap_path):
        download_dataset(dataset_name="meta", data_path=data_path, redownload=True)

    sample_info = pd.read_csv(depmap_path, dtype=str, low_memory=False)

    merged = _harmonize_disease_annotations(df_cellosaurus, sample_info)

    # Synonym mapping

    tissue_lookup = {syn.lower(): tissue for tissue, syns in _tissue_synonyms.items() for syn in syns}

    # Map tissues
    merged["disease_cleaned"] = merged["disease_combined"].map(tissue_lookup).fillna("Unknown").str.title()

    # Final tissue map
    final = merged[
        [
            "cellosaurus_id",
            "cell_line_name",
            "DepMap_ID",
            "disease",
            "disease_combined",
            "disease_cleaned",
            "disease_sutype",
            "disease_sub_subtype",
            "culture_type",
            "culture_medium",
            "gender",
            "source",
            "cellosaurus_derived_from_site",
            "cellosaurus_disease",
        ]
    ]

    # Make sure the mapping has unique index entries
    tissue_map = final.drop_duplicates(subset="cellosaurus_id").set_index("cellosaurus_id")["disease_cleaned"]

    tissue_map = _apply_manual_cell_line_corrections(tissue_map)

    final.loc[:, "tissue"] = final.loc[:, "cellosaurus_id"].map(tissue_map)
    final = final.copy()
    if save_tissue_mapping:
        final.drop_duplicates(subset="cellosaurus_id", inplace=True)
        tissue_mapping_path = os.path.join(data_path, "meta", "tissue_mapping.csv")
        final.to_csv(tissue_mapping_path, index=False)

    # Add tissue column to each dataset
    for ds in datasets:
        path = os.path.join(data_path, ds, f"{ds}.csv")
        if not os.path.exists(path):
            print(f"Dataset {path} not found, skipping.")
            continue

        df = pd.read_csv(path, low_memory=False)

        df["tissue"] = df["cellosaurus_id"].map(tissue_map)

        df.to_csv(path, index=False)


if __name__ == "__main__":
    main()



"""
Contains all function required for CurveCurator fitting.

CurveCurator publication:
Bayer, F.P., Gander, M., Kuster, B. et al. CurveCurator: a recalibrated F-statistic to assess,
classify, and explore significance of dose–response curves. Nat Commun 14, 7902 (2023).
https://doi-org.eaccess.tum.edu/10.1038/s41467-023-43696-z

CurveCurator applies a recalibrated F-statistic for p-value estimation of 4-point log-logistic
regression fits. In drevalpy, this can be used to generate training data with higher quality, since
quality measures, such as p-value, R2, or relevance score can be used to filter out viability
measurements of low quality.
"""

import subprocess
from pathlib import Path

import numpy as np
import pandas as pd
import toml

from ..pipeline_function import pipeline_function


def _prepare_raw_data(curve_df: pd.DataFrame, output_dir: Path, prefix: str = ""):
    if "replicate" in curve_df.columns:
        n_replicates = curve_df["replicate"].nunique()
        pivot_columns = ["dose", "replicate"]
    else:
        n_replicates = 1
        pivot_columns = ["dose"]

    df = curve_df.pivot(index=["sample", "drug"], columns=pivot_columns, values="response")

    if "replicate" in curve_df.columns:
        control_df = pd.DataFrame({(0.0, col_id): 1.0 for col_id in range(n_replicates)}, index=df.index)
    else:
        control_df = pd.DataFrame({0.0: 1.0}, index=df.index)

    df = pd.concat([control_df, df], axis=1)

    concentrations = df.columns.sort_values()
    doses = concentrations.get_level_values(0).to_list()
    df = df[concentrations]

    experiments = np.arange(df.shape[1])
    df.insert(0, "Name", ["|".join(map(str, i)) for i in df.index.tolist()])
    df.reset_index(drop=True)

    df.columns = ["Name"] + [f"Raw {i}" for i in experiments]

    curvecurator_folder = output_dir / prefix
    curvecurator_folder.mkdir(exist_ok=True, parents=True)
    df.to_csv(curvecurator_folder / "curvecurator_input.tsv", sep="\t", index=False)

    return len(experiments), doses, n_replicates, len(df)


def _prepare_toml(
    filename: str, n_exp: int, n_replicates: int, doses: list[float], dataset_name: str, cores: int, condition: str = ""
):
    config = {
        "Meta": {
            "id": filename,
            "description": dataset_name,
            "condition": condition,
            "treatment_time": "72 h",
        },
        "Experiment": {
            "experiments": range(n_exp),
            "doses": doses,
            "dose_scale": "1e-06",
            "dose_unit": "M",
            "control_experiment": [i for i in range(n_replicates)],
            "measurement_type": "OTHER",
            "data_type": "OTHER",
            "search_engine": "OTHER",
            "search_engine_version": "0",
        },
        "Paths": {
            "input_file": "curvecurator_input.tsv",
            "curves_file": "curves.tsv",
            "normalization_file": "norm.txt",
            "mad_file": "mad.txt",
            "dashboard": "dashboard.html",
        },
        "Processing": {
            "available_cores": cores,
            "max_missing": max(len(doses) - 5, 0),
            "imputation": False,
            "normalization": False,
        },
        "Curve Fit": {
            "type": "OLS",
            "speed": "exhaustive",
            "max_iterations": 1000,
            "interpolation": False,
            "control_fold_change": True,
        },
        "F Statistic": {
            "optimized_dofs": True,
            "alpha": 0.05,
            "fc_lim": 0.45,
        },
    }
    return config


def _exec_curvecurator(output_dir: Path, batched: bool = True):
    """
    Execute CurveCurator in batch mode.

    This function spawns a subprocess that runs CurveCurator for all config.toml files that
    are listed in a file "configlist.txt" in the provided output directory.

    :param output_dir: The directory containing einter configlist.txt as well as subfolders for
        all the paths listed in configlist.txt that function as input and output directories for
        batched CurveCurator execution, or the directory containig a single config.toml and
        corresponding viability input.
    :param batched: If True, run CurveCurator in batched mode (default), iterating over a list
        of configs spefified in <output_dir>/configlist.txt and consecutively executing each
        CurveCurator run. If False, run a single CurveCurator run (this can be used for
        parallelisation).
    """
    if batched:
        command = ["CurveCurator", str(output_dir / "configlist.txt"), "--mad", "--batch"]
    else:
        command = ["CurveCurator", str(output_dir / "config.toml"), "--mad"]
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    process.communicate()


def _calc_ic50(model_params_df: pd.DataFrame):
    """
    Calculate the IC50 in M from a fitted model.

    This function expects a dataframe that was processed in the postprocess function, containing
    the columns "Front", "Back", "Slope", "pEC50". It calculates the IC50 for all the models in the
    dataframe in closed form and adds the column IC50_curvecurator to the input dataframe.

    :param model_params_df: a dataframe containing the fitted parameters
    """

    def ic50(front, back, slope, pec50):
        with np.errstate(invalid="ignore"):
            return np.power(10, (np.log10((front - 0.5) / (0.5 - back)) - slope * pec50) / slope)

    front = model_params_df["Front"].values
    back = model_params_df["Back"].values
    slope = model_params_df["Slope"].values
    pec50 = model_params_df["pEC50_curvecurator"].values

    model_params_df["IC50_curvecurator"] = ic50(front, back, slope, pec50)


@pipeline_function
def preprocess(input_file: str | Path, output_dir: str | Path, dataset_name: str, cores: int):
    """
    Preprocess raw viability data and create required input files for CurveCurator.

    This function takes an input file containing raw viability in long format. The required columns
    are "dose", "response", "sample", and "drug", with an optional "replicate" column.
    If there are multiple dose ranges or numbers of replicates, groups in the form
    (maxdose, mindose, n_replicates) are created to keep the number of parameters for fitting low
    and the input dataframes for curvecurator as dense as possible.
    All dosages must be provided in µM!
    All responses must be normalized against the control already without the response for the control.

    :param input_file: Path to csv file containing the raw viability data
    :param output_dir: Path to store all the files to, including the preprocessed data, the config.toml
        for CurveCurator, CurveCurator's output files, and the postprocessed data
    :param dataset_name: Name of the dataset
    :param cores: The number of cores to be used for fitting the curves using CurveCurator.
        This parameter is written into the config.toml, but it is min of the number of curves to fit
        and the number given (min(n_curves, cores))
    :raises ValueError: If required columns are not found in the provided input file.
    """
    input_file = Path(input_file)
    output_dir = Path(output_dir)
    required_columns = ["dose", "response", "sample", "drug", "replicate"]
    converters = {"dose": float, "response": float, "sample": str, "drug": str, "replicate": int}
    try:
        curve_df = pd.read_csv(input_file, usecols=required_columns, converters=converters)
    except ValueError:
        required_columns.pop()
        del converters["replicate"]
        curve_df = pd.read_csv(input_file, usecols=required_columns, converters=converters)

    if not all([col in curve_df.columns for col in required_columns]):
        raise ValueError(f"Missing columns in viability data. Required columns are {required_columns}.")
    groupby = []

    curve_df["mindose"] = curve_df.groupby(["sample", "drug"], as_index=False)["dose"].transform("min")
    curve_df["maxdose"] = curve_df.groupby(["sample", "drug"], as_index=False)["dose"].transform("max")

    if curve_df["maxdose"].nunique() > 1:
        groupby.append("maxdose")
    if curve_df["mindose"].nunique() > 1:
        groupby.append("mindose")
    if "replicate" in curve_df.columns:
        curve_df["nreplicates"] = curve_df.groupby(["sample", "drug"])["replicate"].transform("nunique")
        if curve_df["nreplicates"].nunique() > 1:
            groupby.append("nreplicates")

    if len(groupby) > 0:
        drug_df_groups = curve_df.groupby(groupby)
    else:
        drug_df_groups = [("drug_treatment", curve_df)]

    configs = []

    for index, df in drug_df_groups:
        prefix = "_".join([f"{s}" for s in index])
        n_exp, doses, n_replicates, n_curves_to_fit = _prepare_raw_data(
            curve_df=df, output_dir=output_dir, prefix=prefix
        )
        config = _prepare_toml(
            input_file.name, n_exp, n_replicates, doses, dataset_name, min(n_curves_to_fit, cores), prefix
        )
        config_path = output_dir / prefix / "config.toml"
        with open(config_path, "w") as f:
            toml.dump(config, f)
        configs.append(f"{config_path}\n")

    with open(output_dir / "configlist.txt", "w") as f:
        f.writelines(configs)


@pipeline_function
def postprocess(output_folder: str | Path, dataset_name: str):
    """
    Postprocess CurveCurator output files.

    This function reads all curves.tsv files created by CurveCurator, which contain the
    fitted curve parameters, postprocesses them to be used by drevalpy and combines everything
    in one <dataset_name>.csv file for usage by drevalpy.

    :param output_folder: Path to the output folder of CurveCurator containing the curves.txt file.
    :param dataset_name: The name of the dataset, will be used to prepend the postprocessed <dataset_name>.csv file
    """
    output_folder = Path(output_folder)
    curvecurator_output_files = output_folder.rglob("curves.tsv")
    required_columns = {
        "Name": "Name",
        "pEC50": "pEC50_curvecurator",
        "pEC50 Error": "pEC50Error",
        "Curve Slope": "Slope",
        "Curve Front": "Front",
        "Curve Back": "Back",
        "Curve Fold Change": "FoldChange",
        "Curve AUC": "AUC_curvecurator",
        "Curve R2": "R2",
        "Curve P_Value": "pValue",
        "Curve Relevance Score": "RelevanceScore",
        "Curve F_Value": "fValue",
        "Curve Log P_Value": "negLog10pValue",
        "Signal Quality": "SignalQuality",
        "Curve RMSE": "RMSE",
        "Curve F_Value SAM Corrected": "fValueSAMCorrected",
        "Curve Regulation": "Regulation",
    }

    with open(output_folder / f"{dataset_name}.csv", "w") as f:
        first_file = True
        for output_file in curvecurator_output_files:
            fitted_curve_data = pd.read_csv(output_file, sep="\t", usecols=required_columns).rename(
                columns=required_columns
            )
            fitted_curve_data[["cell_line_id", "drug_id"]] = fitted_curve_data.Name.str.split("|", expand=True)
            fitted_curve_data["EC50_curvecurator"] = np.power(
                10, -fitted_curve_data["pEC50_curvecurator"].values
            )  # in CurveCurator 10^-pEC50 = EC50
            _calc_ic50(fitted_curve_data)
            fitted_curve_data.to_csv(f, index=None, header=first_file, mode="a")
            first_file = False
        f.close()


def fit_curves(input_file: str | Path, output_dir: str | Path, dataset_name: str, cores: int):
    """
    Fit curves for provided raw viability data.

    This functions reads viability data in a predefined input format, preprocesses the data
    to be readable by CurveCurator, fits curves to the data using CurveCurator, and postprocesses
    the fitted data to a format required by drevalpy.

    :param input_file: Path to the file containing the raw viability data
    :param output_dir: Path to store all the files to, including the preprocessed data, the config.toml
        for CurveCurator, CurveCurator's output files, and the postprocessed data
    :param dataset_name: The name of the dataset, will be used to prepend the postprocessed <dataset_name>.csv file
    :param cores: The number of cores to be used for fitting the curves using CurveCurator.
        This parameter is written into the config.toml, but it is min of the number of curves to fit
        and the number given (min(n_curves, cores))
    """
    preprocess(input_file, output_dir, dataset_name, cores)
    _exec_curvecurator(Path(output_dir))
    postprocess(output_dir, dataset_name)



"""Utility functions for the visualization part of the package."""

import os
import pathlib
import re
import shutil
from typing import TextIO

import importlib_resources
import numpy as np
import pandas as pd

from ..datasets.dataset import DrugResponseDataset
from ..evaluation import AVAILABLE_METRICS, evaluate
from ..pipeline_function import pipeline_function
from . import (
    ComparisonScatter,
    CriticalDifferencePlot,
    CrossStudyTables,
    Heatmap,
    RegressionSliderPlot,
    VioHeat,
    Violin,
)


def create_output_directories(result_path: pathlib.Path, custom_id: str) -> None:
    """
    If they do not exist yet, make directories for the visualization files.

    :param result_path: path to the results
    :param custom_id: run id passed via command line
    """
    for dir in [
        "violin_plots",
        "heatmaps",
        "regression_plots",
        "comp_scatter",
        "html_tables",
        "critical_difference_plots",
    ]:
        os.makedirs(pathlib.Path(result_path / custom_id / dir), exist_ok=True)


def _parse_layout(f: TextIO, path_to_layout: str, test_mode: str) -> None:
    """
    Parse the layout file and write it to the output file.

    :param f: file to write to
    :param path_to_layout: path to the layout file
    :param test_mode: test mode, e.g., LPO
    """
    with open(path_to_layout, encoding="utf-8") as layout_f:
        layout = layout_f.readlines()
    if path_to_layout.endswith("index_layout.html"):
        # remove the last 2 lines (</body>, </html>)
        layout = layout[:-2]
    else:
        # remove the last 3 lines (</div>, </body>, </html>)
        layout = layout[:-3]
        # replace LPOLCOLDO with the test mode
        layout = [line.replace("LPOLCOLDO", test_mode) for line in layout]
    f.write("".join(layout))


def parse_results(path_to_results: str, dataset: str) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Parse the results from the given directory.

    :param path_to_results: path to the results directory
    :param dataset: dataset name, e.g., GDSC2
    :returns: evaluation results, evaluation results per drug, evaluation results per cell line, and true vs. predicted
        values
    """
    print("Generating result tables ...")
    # generate list of all result files
    result_dir = pathlib.Path(path_to_results)
    result_files = list(result_dir.rglob("*.csv"))
    # filter for all files that follow this pattern:
    # result_dir/*/{predictions|cross_study|randomization|robustness}/*.csv
    # Convert the path to a forward-slash version for the regex (for Windows)
    result_dir_str = str(result_dir).replace("\\", "/")
    pattern = re.compile(
        rf"{result_dir_str}/{dataset}/"
        r"(LPO|LCO|LDO|LTO)/[^/]+/(predictions|cross_study|randomization|robustness)/.*\.csv$"
    )
    result_files = [file for file in result_files if pattern.match(str(file).replace("\\", "/"))]

    # inititalize dictionaries to store the evaluation results
    evaluation_results = None
    evaluation_results_per_drug = None
    evaluation_results_per_cell_line = None
    true_vs_pred = None

    # read every result file and compute the evaluation metrics
    for file in result_files:
        rel_file = str(os.path.normpath(file.relative_to(result_dir))).replace("\\", "/")
        print(f'Evaluating file: "{rel_file}" ...')
        file_parts = rel_file.split("/")
        dataset = file_parts[0]
        test_mode = file_parts[1]
        algorithm = file_parts[2]
        (
            overall_eval,
            eval_results_per_drug,
            eval_results_per_cl,
            t_vs_p,
            model_name,
        ) = evaluate_file(pred_file=file, test_mode=test_mode, model_name=algorithm)

        evaluation_results = (
            overall_eval if evaluation_results is None else pd.concat([evaluation_results, overall_eval])
        )
        true_vs_pred = t_vs_p if true_vs_pred is None else pd.concat([true_vs_pred, t_vs_p])

        if eval_results_per_drug is not None:
            evaluation_results_per_drug = (
                eval_results_per_drug
                if evaluation_results_per_drug is None
                else pd.concat([evaluation_results_per_drug, eval_results_per_drug])
            )

        if eval_results_per_cl is not None:
            evaluation_results_per_cell_line = (
                eval_results_per_cl
                if evaluation_results_per_cell_line is None
                else pd.concat([evaluation_results_per_cell_line, eval_results_per_cl])
            )

    return (
        evaluation_results,
        evaluation_results_per_drug,
        evaluation_results_per_cell_line,
        true_vs_pred,
    )


@pipeline_function
def evaluate_file(
    pred_file: pathlib.Path, test_mode: str, model_name: str
) -> tuple[pd.DataFrame, pd.DataFrame | None, pd.DataFrame | None, pd.DataFrame, str]:
    """
    Evaluate the predictions from the final models.

    :param pred_file: path to the prediction file
    :param test_mode: test mode, e.g., LPO
    :param model_name: model name, e.g., SimpleNeuralNetwork
    :return: evaluation results, evaluation results per drug, evaluation results per cell line, true vs. predicted
        values, and model name
    """
    print("Parsing file:", os.path.normpath(pred_file))
    dataset = DrugResponseDataset.from_csv(pred_file)

    model = _generate_model_names(test_mode=test_mode, model_name=model_name, pred_file=pred_file)

    # overall evaluation
    overall_eval = {model: evaluate(dataset, list(AVAILABLE_METRICS.keys()))}

    true_vs_pred = pd.DataFrame(
        {
            "model": [model for _ in range(len(dataset.response))],
            "drug": dataset.drug_ids,
            "cell_line": dataset.cell_line_ids,
            "y_true": dataset.response,
            "y_pred": dataset.predictions,
        }
    )

    evaluation_results_per_drug = None
    evaluation_results_per_cl = None

    if "LPO" in model or "LDO" in model:
        evaluation_results_per_drug = _evaluate_per_group(
            df=true_vs_pred,
            group_by="drug",
            eval_results_per_group=evaluation_results_per_drug,
            model=model,
        )
    if "LPO" in model or "LCO" in model or "LTO" in model:
        evaluation_results_per_cl = _evaluate_per_group(
            df=true_vs_pred,
            group_by="cell_line",
            eval_results_per_group=evaluation_results_per_cl,
            model=model,
        )
    overall_eval = pd.DataFrame.from_dict(overall_eval, orient="index")

    return (
        overall_eval,
        evaluation_results_per_drug,
        evaluation_results_per_cl,
        true_vs_pred,
        model,
    )


@pipeline_function
def prep_results(
    eval_results: pd.DataFrame,
    eval_results_per_drug: pd.DataFrame,
    eval_results_per_cell_line: pd.DataFrame,
    t_vs_p: pd.DataFrame,
    path_data: pathlib.Path,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Prepare the results by introducing new columns for algorithm, randomization, test_mode, split, CV_split.

    :param eval_results: evaluation results
    :param eval_results_per_drug: evaluation results per drug
    :param eval_results_per_cell_line: evaluation results per cell line
    :param t_vs_p: true vs. predicted values
    :param path_data: path to the data
    :returns: the same dataframes with new columns
    """
    # get metadata
    print("Getting information about drugs and cell lines ...")
    drug_metadata: dict[str, str] = dict()
    cell_line_metadata: dict[str, str] = dict()
    for root, _, files in os.walk(path_data):
        for file in files:
            if file == "drug_names.csv":
                drug_names = pd.read_csv(os.path.join(root, file), index_col=0)
                # make index to str
                drug_names.index = drug_names.index.astype(str)
                # index: pubchem_id, column: drug_name
                drug_metadata.update(zip(drug_names.index, drug_names["drug_name"]))
            elif file == "cell_line_names.csv":
                cell_line_names = pd.read_csv(os.path.join(root, file), index_col=0)
                # index: cellosaurus_id, column: cell_line_name
                cell_line_metadata.update(zip(cell_line_names["cell_line_name"], cell_line_names.index))

    # add variables
    # split the index by "_" into: algorithm, randomization, test_mode, split, CV_split
    print("Reformatting the evaluation results ...")
    new_columns = eval_results.index.str.split("_", expand=True).to_frame()
    new_columns.columns = [
        "algorithm",
        "rand_setting",
        "test_mode",
        "split",
        "CV_split",
    ]
    new_columns.index = eval_results.index
    eval_results = pd.concat([new_columns.drop("split", axis=1), eval_results], axis=1)
    if eval_results_per_drug is not None:
        print("Reformatting the evaluation results per drug ...")
        eval_results_per_drug[["algorithm", "rand_setting", "test_mode", "split", "CV_split"]] = eval_results_per_drug[
            "model"
        ].str.split("_", expand=True)
        all_drugs = [drug_metadata[drug] for drug in eval_results_per_drug["drug"]]
        eval_results_per_drug["drug_name"] = all_drugs
        # rename drug to pubchem_id
        eval_results_per_drug = eval_results_per_drug.rename(columns={"drug": "pubchem_id"})
    if eval_results_per_cell_line is not None:
        print("Reformatting the evaluation results per cell line ...")
        eval_results_per_cell_line[["algorithm", "rand_setting", "test_mode", "split", "CV_split"]] = (
            eval_results_per_cell_line["model"].str.split("_", expand=True)
        )
        all_cello_ids = [cell_line_metadata[cell_line] for cell_line in eval_results_per_cell_line["cell_line"]]
        eval_results_per_cell_line["cellosaurus_id"] = all_cello_ids
        eval_results_per_cell_line = eval_results_per_cell_line.rename(columns={"cell_line": "cell_line_name"})

    print("Reformatting the true vs. predicted values ...")
    t_vs_p[["algorithm", "rand_setting", "test_mode", "split", "CV_split"]] = t_vs_p["model"].str.split(
        "_", expand=True
    )
    t_vs_p = t_vs_p.drop("split", axis=1)
    all_drugs = [drug_metadata[drug] for drug in t_vs_p["drug"]]
    t_vs_p["drug_name"] = all_drugs
    all_cello_ids = [cell_line_metadata[cell_line] for cell_line in t_vs_p["cell_line"]]
    t_vs_p["cellosaurus_id"] = all_cello_ids
    t_vs_p = t_vs_p.rename(columns={"cell_line": "cell_line_name", "drug": "pubchem_id"})
    t_vs_p["pubchem_id"] = t_vs_p["pubchem_id"].astype(str)

    if "NaiveMeanEffectsPredictor" in eval_results["algorithm"].unique():
        eval_results = _normalize_metrics_by_mean_effects(
            evaluation_results=eval_results,
            true_vs_pred=t_vs_p,
        )

    return (
        eval_results,
        eval_results_per_drug,
        eval_results_per_cell_line,
        t_vs_p,
    )


def _normalize_metrics_by_mean_effects(
    evaluation_results: pd.DataFrame,
    true_vs_pred: pd.DataFrame,
) -> pd.DataFrame:
    """
    Normalize the y_true and y_pred values by the predictions of the NaiveMeanEffectsPredictor.

    Then recalculate the metrics.
    :param evaluation_results: results of the evaluation
    :param true_vs_pred: all true vs. predicted values
    :return: modified evaluation results
    """
    eval_results_mod = {}
    naive_mean_effects_dict = {}
    for rand_setting in evaluation_results["rand_setting"].unique():
        for test_mode in evaluation_results["test_mode"].unique():
            naive_mean_effects_dict[f"{test_mode}_{rand_setting}"] = true_vs_pred[
                (true_vs_pred["algorithm"] == "NaiveMeanEffectsPredictor")
                & (true_vs_pred["rand_setting"] == rand_setting)
                & (true_vs_pred["test_mode"] == test_mode)
            ]
    # do this: per algorithm, per rand test_mode, per test_mode, per CV split
    for algorithm in evaluation_results["algorithm"].unique():
        for rand_setting in evaluation_results["rand_setting"].unique():
            for test_mode in evaluation_results["test_mode"].unique():
                print(f"Calculating normalized metrics for {algorithm}, {rand_setting}, " f"{test_mode} ...")
                setting_subset = true_vs_pred[
                    (true_vs_pred["algorithm"] == algorithm)
                    & (true_vs_pred["rand_setting"] == rand_setting)
                    & (true_vs_pred["test_mode"] == test_mode)
                ]
                if setting_subset.empty:
                    continue
                naive_mean_effects = naive_mean_effects_dict[f"{test_mode}_{rand_setting}"]
                naive_mean_effects = naive_mean_effects[["drug_name", "cell_line_name", "CV_split", "y_pred"]]
                naive_mean_effects = naive_mean_effects.rename(columns={"y_pred": "y_pred_naive"})
                setting_subset = setting_subset[["drug_name", "cell_line_name", "CV_split", "y_true", "y_pred"]]
                setting_subset = setting_subset.merge(
                    naive_mean_effects, on=["drug_name", "cell_line_name", "CV_split"], how="left"
                )
                setting_subset["y_true"] = setting_subset["y_true"] - setting_subset["y_pred_naive"]
                setting_subset["y_pred"] = setting_subset["y_pred"] - setting_subset["y_pred_naive"]
                for cv_split in setting_subset["CV_split"].unique():
                    setting_subset_cv = setting_subset[setting_subset["CV_split"] == cv_split]
                    dt = DrugResponseDataset(
                        response=setting_subset_cv["y_true"].to_numpy(),
                        cell_line_ids=setting_subset_cv["cell_line_name"].to_numpy(),
                        drug_ids=setting_subset_cv["drug_name"].to_numpy(),
                        predictions=setting_subset_cv["y_pred"].to_numpy(),
                    )
                    res = evaluate(
                        dataset=dt,
                        metric=list(AVAILABLE_METRICS.keys() - {"MAE", "MSE", "RMSE"}),
                    )
                    eval_results_mod[f"{algorithm}_{rand_setting}_{test_mode}_split_{cv_split}"] = res
    mod_table = pd.DataFrame.from_dict(eval_results_mod, orient="index")
    mod_table.columns = [f"{col}: normalized" for col in mod_table.columns]
    evaluation_results = evaluation_results.merge(mod_table, left_index=True, right_index=True)
    return evaluation_results


def _generate_model_names(test_mode: str, model_name: str, pred_file: pathlib.Path) -> str:
    """
    Generate the model names based on the prediction file.

    :param test_mode: test mode, e.g., LPO
    :param model_name: model name, e.g., SimpleNeuralNetwork
    :param pred_file: file containing the predictions
    :returns: unique name of run = {model_name}_{pred_setting}_{test_mode}_{split}
    :raises ValueError: if the prediction test_mode is unknown
    """
    file_parts = os.path.basename(pred_file).split("_")
    pred_rand_rob = file_parts[0]
    if pred_rand_rob == "predictions":
        pred_setting = "predictions"
    elif pred_rand_rob == "randomization":
        pred_setting = "randomize-" + "-".join(file_parts[1:-2])
    elif pred_rand_rob == "robustness":
        pred_setting = "-".join(file_parts[:2])
    elif pred_rand_rob == "cross":
        pred_setting = "cross-study-" + file_parts[2]
    else:
        raise ValueError(f"Unknown prediction test_mode: {pred_rand_rob}")
    split = "_".join(os.path.basename(pred_file).split(".")[0].split("_")[-2:])
    return f"{model_name}_{pred_setting}_{test_mode}_{split}"


def _evaluate_per_group(
    df: pd.DataFrame,
    group_by: str,
    eval_results_per_group: pd.DataFrame | None,
    model: str,
) -> pd.DataFrame:
    """
    Evaluate the predictions per group.

    :param df: true vs. predicted values
    :param group_by: cell line or drug
    :param eval_results_per_group: evaluation results per group
    :param model: model name
    :returns: dictionary with the normalized group evaluation results and the evaluation results per group
    """
    # calculate the mean of y_true per drug
    print(f"Calculating {group_by}-wise evaluation measures …")
    # evaluation per group
    eval_results_per_group = compute_evaluation(df, eval_results_per_group, group_by, model)
    return eval_results_per_group


def compute_evaluation(df: pd.DataFrame, return_df: pd.DataFrame | None, group_by: str, model: str) -> pd.DataFrame:
    """
    Compute the evaluation metrics per group.

    :param df: true vs. predicted values with mean_y_true_per_{group_by} column
    :param return_df: DataFrame to store the results
    :param group_by: either cell line or drug
    :param model: model name
    :returns: dataframe with the evaluation results per group
    """
    result_per_group = df.groupby(group_by)[["y_true", "cell_line", "drug", "y_pred"]].apply(
        lambda x: evaluate(
            DrugResponseDataset(
                response=x["y_true"].to_numpy(),
                cell_line_ids=x["cell_line"].to_numpy(),
                drug_ids=x["drug"].to_numpy(),
                predictions=x["y_pred"].to_numpy(),
            ),
            list(AVAILABLE_METRICS.keys()),
        )
    )
    groups = result_per_group.index
    result_per_group = pd.json_normalize(result_per_group)
    result_per_group[group_by] = groups
    result_per_group["model"] = model
    if return_df is None:
        return_df = pd.DataFrame(result_per_group)
    else:
        return_df = pd.concat([return_df, result_per_group])
    return return_df


@pipeline_function
def write_results(
    path_out: str,
    eval_results: pd.DataFrame,
    eval_results_per_drug: pd.DataFrame,
    eval_results_per_cl: pd.DataFrame,
    t_vs_p: pd.DataFrame,
) -> None:
    """
    Write the results to csv files.

    :param path_out: path to the output directory, e.g., results/my_run/
    :param eval_results: evaluation results
    :param eval_results_per_drug: evaluation results per drug
    :param eval_results_per_cl: evaluation results per cell line
    :param t_vs_p: true vs. predicted values
    """
    eval_results.to_csv(f"{path_out}evaluation_results.csv", index=True)
    if eval_results_per_drug is not None:
        eval_results_per_drug.to_csv(f"{path_out}evaluation_results_per_drug.csv", index=True)
    if eval_results_per_cl is not None:
        eval_results_per_cl.to_csv(f"{path_out}evaluation_results_per_cl.csv", index=True)
    t_vs_p.to_csv(f"{path_out}true_vs_pred.csv", index=True)


@pipeline_function
def create_index_html(custom_id: str, test_modes: list[str], prefix_results: str) -> None:
    """
    Create the index.html file.

    :param custom_id: custom id for the results, e.g., my_run
    :param test_modes: list of test modes, e.g., ["LPO", "LCO", "LDO"]
    :param prefix_results: path to the results directory, e.g., results/my_run
    """
    # copy images to the results directory
    file_to_copy = [
        "favicon.png",
        "nf-core-drugresponseeval_logo_light.png",
    ]
    for file in file_to_copy:
        file_path = os.path.join(
            str(importlib_resources.files("drevalpy")),
            "visualization",
            "style_utils",
            file,
        )
        shutil.copyfile(file_path, os.path.join(prefix_results, file))

    layout_path = os.path.join(
        str(importlib_resources.files("drevalpy")),
        "visualization",
        "style_utils",
        "index_layout.html",
    )
    idx_html_path = os.path.join(prefix_results, "index.html")
    with open(idx_html_path, "w", encoding="utf-8") as f:
        _parse_layout(f=f, path_to_layout=layout_path, test_mode="")
        f.write('<div class="main">\n')
        f.write('<img src="nf-core-drugresponseeval_logo_light.png" ' 'width="364px" height="100px" alt="Logo">\n')
        f.write(f"<h1>Results for {custom_id}</h1>\n")
        f.write("<h2>Available settings</h2>\n")
        f.write('<div style="display: inline-block;">\n')
        f.write("<p>Click on the images to open the respective report in a new tab.</p>\n")

        test_modes.sort()
        for test_mode in test_modes:
            img_path = os.path.join(
                str(importlib_resources.files("drevalpy")),
                "visualization",
                "style_utils",
                f"{test_mode}.png",
            )
            shutil.copyfile(img_path, os.path.join(prefix_results, f"{test_mode}.png"))
            f.write(
                f'<a href="{test_mode}.html" target="_blank"><img src="{test_mode}.png" '
                f'style="width:300px;height:300px;"></a>\n'
            )
        f.write("</div>\n")
        f.write("</div>\n")
        f.write("</body>\n")
        f.write("</html>\n")


@pipeline_function
def create_html(run_id: str, test_mode: str, files: list, prefix_results: str) -> None:
    """
    Create the html file for the given test mode, e.g., LPO.html.

    :param run_id: custom id for the results, e.g., my_run
    :param test_mode: test mode, e.g., LPO
    :param files: list of files in the results directory
    :param prefix_results: path to the results directory, e.g., results/my_run
    """
    page_layout = os.path.join(
        str(importlib_resources.files("drevalpy")),
        "visualization/style_utils/page_layout.html",
    )
    html_path = os.path.join(prefix_results, f"{test_mode}.html")

    with open(html_path, "w", encoding="utf-8") as f:
        _parse_layout(f=f, path_to_layout=page_layout, test_mode=test_mode)
        f.write(f"<h1>Results for {run_id}: {test_mode}</h1>\n")

        # Critical difference plot
        f = CriticalDifferencePlot.write_to_html(test_mode=test_mode, f=f)

        # Violin plots
        f = VioHeat.write_to_html(test_mode=test_mode, f=f, files=files, plot="Violin")

        # Heatmaps
        f = VioHeat.write_to_html(test_mode=test_mode, f=f, files=files, plot="Heatmap")

        # Regression plots
        f = RegressionSliderPlot.write_to_html(test_mode=test_mode, f=f, files=files)

        # Correlation comparison: Drug
        f = ComparisonScatter.write_to_html(test_mode=test_mode, f=f, files=files)

        # Cross-study evaluation tables
        f = CrossStudyTables.write_to_html(test_mode=test_mode, f=f, files=files, prefix=prefix_results)

        f.write("</div>\n")
        f.write("</body>\n")
        f.write("</html>\n")


def draw_test_mode_plots(
    test_mode: str,
    ev_res: pd.DataFrame,
    ev_res_per_drug: pd.DataFrame | None,
    ev_res_per_cell_line: pd.DataFrame | None,
    custom_id: str,
    path_data: pathlib.Path,
    result_path: pathlib.Path,
) -> np.ndarray:
    """
    Draw all plots for a specific test_mode (LPO, LCO, LDO, LTO).

    :param test_mode: test_mode
    :param ev_res: overall evaluation results
    :param ev_res_per_drug: evaluation results per drug
    :param ev_res_per_cell_line: evaluation results per cell line
    :param custom_id: run id passed via command line
    :param path_data: path to the data
    :param result_path: path to the results
    :returns: list of unique algorithms
    """
    ev_res_subset = ev_res[ev_res["test_mode"] == test_mode]

    # only draw figures for 'real' predictions comparing all models
    eval_results_preds = ev_res_subset[ev_res_subset["rand_setting"] == "predictions"]

    # PIPELINE: DRAW_CRITICAL_DIFFERENCE
    cd_plot = CriticalDifferencePlot(eval_results_preds=eval_results_preds, metric="MSE")
    cd_plot.draw_and_save(
        out_prefix=f"{result_path}/{custom_id}/critical_difference_plots/",
        out_suffix=test_mode,
    )
    # PIPELINE: DRAW_VIOLIN_AND_HEATMAP
    for plt_type in ["violinplot", "heatmap"]:
        if plt_type == "violinplot":
            out_dir = "violin_plots"
        else:
            out_dir = "heatmaps"
        for normalized in [False, True]:
            if normalized:
                out_suffix = f"algorithms_{test_mode}_normalized"
            else:
                out_suffix = f"algorithms_{test_mode}"
            if plt_type == "violinplot":
                out_plot = Violin(
                    df=eval_results_preds,
                    normalized_metrics=normalized,
                    whole_name=False,
                )

            else:
                out_plot = Heatmap(
                    df=eval_results_preds,
                    normalized_metrics=normalized,
                    whole_name=False,
                )
            out_plot.draw_and_save(
                out_prefix=f"{result_path}/{custom_id}/{out_dir}/",
                out_suffix=out_suffix,
            )

    # per group plots
    if test_mode in ("LPO", "LDO"):
        _draw_per_grouping_setting_plots(
            grouping="drug_name",
            ev_res_per_group=ev_res_per_drug,
            test_mode=test_mode,
            custom_id=custom_id,
            result_path=result_path,
        )
    if test_mode in ("LPO", "LCO", "LTO"):
        _draw_per_grouping_setting_plots(
            grouping="cell_line_name",
            ev_res_per_group=ev_res_per_cell_line,
            test_mode=test_mode,
            custom_id=custom_id,
            result_path=result_path,
        )

    # Cross-study evaluation tables
    cross_study_tables = CrossStudyTables(evaluation_metrics=ev_res_subset, path_data=path_data)
    cross_study_tables.draw_and_save(
        out_prefix=f"{result_path}/{custom_id}/html_tables/",
        out_suffix=test_mode,
    )

    return eval_results_preds["algorithm"].unique()


def _draw_per_grouping_setting_plots(
    grouping: str, ev_res_per_group: pd.DataFrame, test_mode: str, custom_id: str, result_path: pathlib.Path
) -> None:
    """
    Draw plots for a specific grouping (drug or cell line) for a specific test_mode (LPO, LCO, LDO).

    :param grouping: drug or cell_line
    :param ev_res_per_group: evaluation results per drug or per cell line
    :param test_mode: test_mode
    :param custom_id: run id passed over command line
    :param result_path: path to the results
    """
    # PIPELINE: DRAW_CORR_COMP
    corr_comp = ComparisonScatter(
        df=ev_res_per_group,
        color_by=grouping,
        test_mode=test_mode,
        algorithm="all",
    )
    if corr_comp.name is not None:
        corr_comp.draw_and_save(
            out_prefix=f"{result_path}/{custom_id}/comp_scatter/",
            out_suffix=corr_comp.name,
        )


def draw_algorithm_plots(
    model: str,
    ev_res: pd.DataFrame,
    ev_res_per_drug: pd.DataFrame | None,
    ev_res_per_cell_line: pd.DataFrame | None,
    t_vs_p: pd.DataFrame,
    test_mode: str,
    custom_id: str,
    result_path: pathlib.Path,
) -> None:
    """
    Draw all plots for a specific algorithm.

    :param model: name of the model/algorithm
    :param ev_res: overall evaluation results
    :param ev_res_per_drug: evaluation results per drug
    :param ev_res_per_cell_line: evaluation results per cell line
    :param t_vs_p: true response values vs. predicted response values
    :param test_mode: test_mode
    :param custom_id: run id passed via command line
    :param result_path: path to the results
    """
    eval_results_algorithm = ev_res[(ev_res["test_mode"] == test_mode) & (ev_res["algorithm"] == model)]
    # PIPELINE: DRAW_VIOLIN_AND_HEATMAP
    for plt_type in ["violinplot", "heatmap"]:
        if len(eval_results_algorithm["rand_setting"].unique()) < 2:
            # only draw plots if there are predictions and another test_mode (randomization/robustness)
            continue
        if plt_type == "violinplot":
            out_dir = "violin_plots"
            out_plot = Violin(
                df=eval_results_algorithm,
                normalized_metrics=False,
                whole_name=True,
            )
        else:
            out_dir = "heatmaps"
            out_plot = Heatmap(
                df=eval_results_algorithm,
                normalized_metrics=False,
                whole_name=True,
            )
        out_plot.draw_and_save(
            out_prefix=f"{result_path}/{custom_id}/{out_dir}/",
            out_suffix=f"{model}_{test_mode}",
        )
    if test_mode in ("LPO", "LDO"):
        _draw_per_grouping_algorithm_plots(
            grouping="drug_name",
            model=model,
            ev_res_per_group=ev_res_per_drug,
            t_v_p=t_vs_p,
            test_mode=test_mode,
            custom_id=custom_id,
            result_path=result_path,
        )
    if test_mode in ("LPO", "LCO", "LTO"):
        _draw_per_grouping_algorithm_plots(
            grouping="cell_line_name",
            model=model,
            ev_res_per_group=ev_res_per_cell_line,
            t_v_p=t_vs_p,
            test_mode=test_mode,
            custom_id=custom_id,
            result_path=result_path,
        )


def _draw_per_grouping_algorithm_plots(
    grouping: str,
    model: str,
    ev_res_per_group: pd.DataFrame,
    t_v_p: pd.DataFrame,
    test_mode: str,
    custom_id: str,
    result_path: pathlib.Path,
):
    """
    Draw plots for a specific grouping (drug or cell line) for a specific algorithm.

    :param grouping: drug or cell_line
    :param model: name of the model/algorithm
    :param ev_res_per_group: evaluation results per drug or per cell line
    :param t_v_p: true response values vs. predicted response values
    :param test_mode: test_mode
    :param custom_id: run id passed via command line
    :param result_path: path to the results
    """
    if len(ev_res_per_group["rand_setting"].unique()) > 1:
        # only draw plots if there are predictions and another test_mode (randomization/robustness)
        # PIPELINE: DRAW_CORR_COMP
        comp_scatter = ComparisonScatter(
            df=ev_res_per_group,
            color_by=grouping,
            test_mode=test_mode,
            algorithm=model,
        )
        if comp_scatter.name is not None:
            comp_scatter.draw_and_save(
                out_prefix=f"{result_path}/{custom_id}/comp_scatter/",
                out_suffix=comp_scatter.name,
            )
    # PIPELINE: DRAW_REGRESSION
    for normalize in [False, True]:
        name_suffix = "_normalized" if normalize else ""
        name = f"{test_mode}_{grouping}{name_suffix}"
        regr_slider = RegressionSliderPlot(
            df=t_v_p,
            test_mode=test_mode,
            model=model,
            group_by=grouping,
            normalize=normalize,
        )
        regr_slider.draw_and_save(
            out_prefix=f"{result_path}/{custom_id}/regression_plots/",
            out_suffix=f"{name}_{model}{name_suffix}",
        )



"""Contains the code needed to draw the correlation comparison scatter plot."""

from io import TextIOWrapper

import pandas as pd
import plotly.graph_objects as go

from ..models import SINGLE_DRUG_MODEL_FACTORY
from ..pipeline_function import pipeline_function
from .outplot import OutPlot


class ComparisonScatter(OutPlot):
    """
    Class to draw scatter plots for comparison of correlation metrics between models.

    Produces two types of plots: an overall comparison plot and a dropdown plot for comparison between all models.
    If one model is consistently better than the other, the points deviate from the identity line (higher if the
    model is on the y-axis, lower if it is on the x-axis.
    The dropdown plot allows to select two models for comparison of their per-drug/per-cell-line pearson correlation.
    The overall plot facets all models and visualizes the density of the points.
    """

    @pipeline_function
    def __init__(
        self,
        df: pd.DataFrame,
        color_by: str,
        test_mode: str,
        metric: str = "R^2",
        algorithm: str = "all",
    ):
        """
        Initialize the ComparisonScatter object.

        :param df: evaluation results per group, either drug or cell line
        :param color_by: group variable, i.e., drug or cell line
        :param test_mode: evaluation test_mode, e.g., LCO (leave-cell-line-out)
        :param metric: correlation metric to be compared. Default is Pearson.
        :param algorithm: used to distinguish between per-algorithm plots and per-test_mode plots (all models then).
        """
        exclude_models = (
            {"NaiveDrugMeanPredictor"}.union({model for model in SINGLE_DRUG_MODEL_FACTORY.keys()})
            if color_by == "drug"
            else {"NaiveCellLineMeanPredictor"}
        )
        exclude_models.add("NaivePredictor")
        exclude_models.add("NaiveMeanEffectsPredictor")

        self.df = df.sort_values("model")
        self.name: str | None = None
        if algorithm == "all":
            # draw plots for comparison between all models
            self.df = self.df[
                (self.df["test_mode"] == test_mode)
                & (self.df["rand_setting"] == "predictions")
                & (~self.df["algorithm"].isin(exclude_models))
                # and exclude all lines for which algorithm starts with any element from
                # exclude_models
                & (~self.df["algorithm"].str.startswith(tuple(exclude_models)))
            ]
            self.name = f"{color_by}_{test_mode}"
        elif algorithm not in exclude_models:
            # draw plots for comparison between all test settings of one model
            self.df = self.df[(self.df["test_mode"] == test_mode) & (self.df["algorithm"] == algorithm)]
            self.name = f"{color_by} {algorithm} {test_mode}"
        if self.df.empty:
            print(f"No data found for {self.name}. Skipping ...")
            return
        self.color_by = color_by
        self.metric = metric

        self.df["test_mode"] = self.df["model"].str.split("_").str[0:3].str.join("_")
        self.models = self.df["test_mode"].unique()

        self.dropdown_fig = go.Figure()
        self.dropdown_buttons_x: list[dict] = list()
        self.dropdown_buttons_y: list[dict] = list()

    @pipeline_function
    def draw_and_save(self, out_prefix: str, out_suffix: str) -> None:
        """
        Draws and saves the scatter plots.

        :param out_prefix: e.g., results/my_run/comp_scatter/
        :param out_suffix: should be self.name
        :raises AssertionError: if out_suffix does not match self.name
        """
        if self.df.empty:
            return
        self._draw()
        if self.name != out_suffix:
            raise AssertionError(f"Name mismatch: {self.name} != {out_suffix}")
        path_out = f"{out_prefix}comp_scatter_{out_suffix}.html"
        self.dropdown_fig.write_html(path_out)

    def _draw(self) -> None:
        """Draws the scatter plots."""
        print("Drawing scatterplots ...")
        self._generate_comp_scatterplots()

        self.dropdown_fig.update_layout(
            title=f'{str(self.color_by).replace("_", " ").capitalize()}-wise scatter plot of {self.metric} '
            f"for each model",
            showlegend=False,
        )
        # Set dropdown menu
        self.dropdown_fig.update_layout(
            updatemenus=[
                {
                    "buttons": self.dropdown_buttons_x,
                    "direction": "down",
                    "showactive": True,
                    "x": 0.0,
                    "xanchor": "left",
                    "y": 1.5,
                    "yanchor": "top",
                },
                {
                    "buttons": self.dropdown_buttons_y,
                    "direction": "down",
                    "showactive": True,
                    "x": 0.5,
                    "xanchor": "left",
                    "y": 1.5,
                    "yanchor": "top",
                },
            ]
        )
        self.dropdown_fig.update_xaxes(range=[-1, 1])
        self.dropdown_fig.update_yaxes(range=[-1, 1])

    @staticmethod
    def write_to_html(test_mode: str, f: TextIOWrapper, *args, **kwargs) -> TextIOWrapper:
        """
        Inserts the generated files into the result HTML file.

        :param test_mode: test_mode, e.g., LCO
        :param f: file to write to
        :param args: unused
        :param kwargs: used to get all files generated by create_report.py / the pipeline
        :returns: the file f
        """
        files: list[str] = kwargs.get("files", [])
        f.write('<h2 id="corr_comp">Comparison of normalized R^2 values</h2>\n')
        f.write(
            "R^2 values can be compared here between models, either per cell line or per drug. "
            "This can either show if a model has consistently higher or lower R^2 values than another model or "
            "identify cell lines/drugs for which models agree or disagree.\n"
            "The x-axis is the first dropdown menu, the y-axis is the second dropdown menu.\n"
        )
        for group_by in ["drug_name", "cell_line_name"]:
            plot_list = [f for f in files if f.startswith("comp_scatter") and f.endswith(f"{test_mode}.html")]
            if f"comp_scatter_{group_by}_{test_mode}.html" in plot_list:
                f.write(f'<h3 id="corr_comp_drug">{group_by.capitalize()}-wise comparison</h3>\n')
                f.write(
                    f'<iframe src="comp_scatter/comp_scatter_{group_by}_{test_mode}.html" '
                    f'width="100%" height="100%" frameBorder="0"></iframe>\n'
                )
                f.write("<h4>Comparisons per model</h4>\n")
                f.write("<ul>\n")
                listed_files = [
                    elem
                    for elem in plot_list
                    if (
                        elem != f"comp_scatter_{group_by}_{test_mode}.html"
                        and elem != f"comp_scatter_overall_{group_by}_{test_mode}.html"
                    )
                ]
                listed_files.sort()
                for group_comparison in listed_files:
                    f.write(
                        f'<li><a href="comp_scatter/{group_comparison}" target="_blank">'
                        f"{group_comparison}</a></li>\n"
                    )
                f.write("</ul>\n")
        return f

    def _generate_comp_scatterplots(self) -> None:
        """Generates the scatter plots."""
        # render first scatterplot that is shown in the dropdown plot
        first_df = self._subset_df(run_id=self.models[0])
        if self.color_by == "drug_name":
            hover_variables = ["drug_name", "pubchem_id"]
        else:
            hover_variables = ["cell_line_name", "cellosaurus_id"]
        scatterplot = go.Scatter(
            x=first_df[self.metric],
            y=first_df[self.metric],
            mode="markers",
            marker=dict(size=6, showscale=False),
            customdata=first_df[hover_variables],
            hovertemplate="<br>".join(
                [
                    f"{self.color_by.capitalize()}: %{{customdata[0]}}",
                    f"{hover_variables[1]}: %{{customdata[1]}}",
                    "x: %{{x:.2f}}",
                    "y: %{{y:.2f}}",
                ]
            ),
            showlegend=True,
            visible=True,
        )
        self.dropdown_fig.add_trace(scatterplot)

        for run_idx in range(len(self.models)):
            run = self.models[run_idx]
            x_df = self._subset_df(run_id=run)
            self.dropdown_buttons_x.append(
                dict(
                    label=run,
                    method="update",
                    args=[
                        {"x": [x_df[self.metric]]},
                        {"xaxis": {"title": run, "range": [-1, 1]}},
                    ],
                )
            )
            for run2_idx in range(len(self.models)):
                run2 = self.models[run2_idx]
                y_df = self._subset_df(run_id=run2)

                # create dropdown buttons for y axis only in the first iteration
                if run_idx == 0:
                    self.dropdown_buttons_y.append(
                        dict(
                            label=run2,
                            method="update",
                            args=[
                                {"y": [y_df[self.metric]]},
                                {"yaxis": {"title": run2, "range": [-1, 1]}},
                            ],
                        )
                    )

    def _subset_df(self, run_id: str) -> pd.DataFrame:
        """
        Subsets the dataframe for a given run_id to the relevant columns and sets the index to the color_by variable.

        :param run_id: user-defined ID of the whole run
        :returns: subsetted dataframe
        """
        subset_cols = [self.metric, self.color_by, "model"]
        if self.color_by == "drug_name":
            subset_cols.append("pubchem_id")
        else:
            subset_cols.append("cellosaurus_id")
        s_df = self.df[self.df["test_mode"] == run_id][subset_cols]
        # sort by color_by variable
        s_df = s_df.sort_values(self.color_by)
        s_df[self.metric] = s_df[self.metric].fillna(0)
        return s_df



"""Plots a heatmap of the evaluation metrics."""

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from ..pipeline_function import pipeline_function
from .vioheat import VioHeat


class Heatmap(VioHeat):
    """Plots a heatmap of the evaluation metrics."""

    @pipeline_function
    def __init__(self, df: pd.DataFrame, normalized_metrics=False, whole_name=False):
        """
        Initialize the Heatmap class.

        :param df: either containing all predictions for all algorithms or all tests for one algorithm (including
            robustness, randomization, … tests then)
        :param normalized_metrics: whether the metrics are normalized
        :param whole_name: whether the whole name should be displayed
        """
        super().__init__(df, normalized_metrics, whole_name)

        self.df = self.df[[col for col in self.df.columns if col in self.all_metrics]]

        self.n_models = len(self.df.index)

        if self.normalized_metrics:
            titles = [
                "Mean R^2: normalized",
                "Mean Correlations: normalized",
            ]
            nr_subplots = 3
            self.plot_settings = ["r2", "correlations"]
        else:
            titles = [
                "Mean R^2",
                "Mean Correlations",
                "Mean Errors",
                "Strictly Standardized Mean Difference for R^2",
                "Strictly Standardized Mean Difference for MSE",
            ]
            self.plot_settings = [
                "r2",
                "correlations",
                "errors",
                "ssmd_R^2",
                "ssmd_MSE",
            ]
            nr_subplots = len(self.plot_settings)

        self.fig = make_subplots(
            rows=nr_subplots,
            cols=1,
            subplot_titles=tuple(titles),
            vertical_spacing=0.1,
        )

    @pipeline_function
    def draw_and_save(self, out_prefix: str, out_suffix: str) -> None:
        """
        Draw the heatmap and save it to a file.

        :param out_prefix: e.g., results/my_run/heatmaps/
        :param out_suffix: e.g., algorithms_normalized
        """
        self._draw()
        path_out = f"{out_prefix}heatmap_{out_suffix}.html"
        self.fig.write_html(path_out)

    def _draw(self) -> None:
        """Draw the heatmap."""
        print("Drawing heatmaps ...")
        for plot_setting in self.plot_settings:

            self._draw_subplots(plot_setting)

        # Dynamically adjust figure height based on number of models
        num_models = self.n_models
        height_per_model = 35  # Increase spacing for each model
        max_height = 5000  # Increase max height if needed
        new_height = min(500 + num_models * height_per_model, max_height)
        self.fig.update_layout(
            height=new_height,
            width=1300,
            title_text="Heatmap of the evaluation metrics",
        )
        self.fig.update_traces(showscale=False)

    def _draw_subplots(self, plot_setting: str) -> None:
        """
        Draw the subplots of the heatmap.

        :param plot_setting: Either  "r2", "correlations", "errors", or "ssmd"
        :raises ValueError: If an unknown plot setting is given
        """
        idx_split = self.df.index.to_series().str.split("_")
        setting = idx_split.str[0:3].str.join("_")
        if plot_setting.startswith("ssmd_"):
            metric_name = plot_setting.split("_")[1]  # Extract metric name (e.g., "ssmd_r2" → "r2")
            dt = self._compute_ssmd(metric_name)
            dt["sort_key"] = dt.max(axis=1)
            dt = dt.sort_values(by="sort_key", ascending=True).drop(columns=["sort_key"])
            dt = dt[dt.index]  # Ensure columns match sorted rows

            if dt.empty:
                print(f"Warning: SSMD heatmap for {metric_name} is empty. Skipping.")
                return
            row_idx = self.plot_settings.index(plot_setting) + 1

            colorscale = "RdBu"
            text_labels = dt.round(3).astype(str)
        else:
            dt_std_errs = self.df.groupby(setting).apply(lambda x: self._calc_summary_metric(x, std_error=True))

            if plot_setting == "r2":
                r2_columns = [col for col in self.df.columns if "R^2" in col]

                dt = self.df[r2_columns].groupby(setting).apply(lambda x: self._calc_summary_metric(x))
                dt = dt.sort_values(by=r2_columns[0], ascending=True)
                dt_std_errs = dt_std_errs[r2_columns]
                dt_std_errs = dt_std_errs.loc[dt.index]

                row_idx = 1
                colorscale = "Blues"
            elif plot_setting == "correlations":
                corr_columns = [
                    col for col in self.df.columns if "Pearson" in col or "Spearman" in col or "Kendall" in col
                ]
                dt = self.df[corr_columns].groupby(setting).apply(lambda x: self._calc_summary_metric(x))
                dt = dt.sort_values(by=corr_columns[0], ascending=True)
                dt_std_errs = dt_std_errs[corr_columns]
                dt_std_errs = dt_std_errs.loc[dt.index]

                row_idx = 2
                colorscale = "Viridis"
            elif plot_setting == "errors":
                error_columns = [col for col in self.df.columns if col in ["MSE", "RMSE", "MAE"]]
                if not error_columns:
                    print("Warning: No error metric columns found. Skipping error heatmap.")
                    return
                dt = self.df[error_columns].groupby(setting).apply(lambda x: self._calc_summary_metric(x))
                dt = dt.sort_values(by=error_columns[0], ascending=False)
                dt_std_errs = dt_std_errs[error_columns]
                dt_std_errs = dt_std_errs.loc[dt.index]

                row_idx = 3
                colorscale = "hot"
            else:
                raise ValueError(f"Unknown plot setting: {plot_setting}")
            text_labels = dt.round(3).astype(str) + " ± " + dt_std_errs.round(3).astype(str)

        labels = [i.replace("_", " ") if self.whole_name else i.split("_")[0] for i in dt.index]
        self.fig.add_trace(
            go.Heatmap(
                z=dt.values,
                x=dt.columns,
                y=labels,
                colorscale=colorscale,
                texttemplate="%{text}",
                text=text_labels,
                textfont={"size": 16},  # size of labels of pixels of the heatmap
            ),
            row=row_idx,
            col=1,
        )

        # Force all y-ticks to be displayed
        self.fig.update_yaxes(
            row=row_idx,
            col=1,
            tickmode="array",
            tickvals=list(range(len(dt.index))),  # Force showing all ticks
            ticktext=labels,
            automargin=True,  # Prevent cutoff
            tickfont=dict(size=15),  # Adjust text size
        )

    def _compute_ssmd(self, metric: str) -> pd.DataFrame:
        """
        Compute Strictly Standardized Mean Difference (SSMD) for a given metric across splits.

        :param metric: The evaluation metric to compute SSMD for (e.g., "R^2", "RMSE", "MAE", "Pearson").
        :return: SSMD heatmap matrix (models × models) as a DataFrame.
        """
        if metric not in self.df.columns:
            print(f"Warning: '{metric}' metric not found in DataFrame. Skipping SSMD heatmap.")
            return pd.DataFrame()

        # Extract only the base model name (remove _predictions_testmode_split_X)
        self.df["model_name"] = self.df.index.to_series().apply(lambda x: x.split("_predictions")[0])

        models = self.df["model_name"].unique()
        ssmd_matrix = pd.DataFrame(index=models, columns=models)

        for m1 in models:
            for m2 in models:
                if m1 == m2:
                    ssmd_matrix.loc[m1, m2] = 0  # No self-comparison
                    continue

                # Get metric values across splits for both models
                values_m1 = self.df[self.df["model_name"] == m1][metric]
                values_m2 = self.df[self.df["model_name"] == m2][metric]

                # Compute SSMD
                mu1, mu2 = values_m1.mean(), values_m2.mean()
                sigma1_sq, sigma2_sq = values_m1.var(ddof=1), values_m2.var(ddof=1)
                ssmd = (mu1 - mu2) / np.sqrt(sigma1_sq + sigma2_sq) if sigma1_sq + sigma2_sq > 0 else np.nan

                ssmd_matrix.loc[m1, m2] = ssmd

        return ssmd_matrix.astype(float)

    @staticmethod
    def _calc_summary_metric(x: pd.DataFrame, std_error: bool = False):
        """
        Calculate the mean or standard error of the metrics.

        :param x: DataFrame containing the metrics
        :param std_error: whether to calculate the standard error or the mean
        :returns: Series containing the mean or standard error of the metrics
        """
        results = pd.Series(index=x.columns)
        for col in x.columns:
            if np.count_nonzero(np.isnan(x[col])) == len(x[col]):
                results[col] = np.nan
            elif std_error:
                results[col] = np.nanstd(x[col]) / np.sqrt(x.shape[0])
            else:
                results[col] = np.nanmean(x[col])
        return results



"""Abstract wrapper class for all visualizations."""

from abc import ABC, abstractmethod
from io import TextIOWrapper


class OutPlot(ABC):
    """Abstract wrapper class for all visualizations."""

    @abstractmethod
    def draw_and_save(self, out_prefix: str, out_suffix: str) -> None:
        """
        Draw and save the plot.

        :param out_prefix: path to output directory for python package
        :param out_suffix: custom suffix for output file
        """
        pass

    @abstractmethod
    def _draw(self) -> None:
        """Draw the plot."""
        pass

    @staticmethod
    @abstractmethod
    def write_to_html(test_mode: str, f: TextIOWrapper, *args, **kwargs) -> TextIOWrapper:
        """
        Write the plot to the final report file.

        :param test_mode: LPO, LCO, LDO
        :param f: the file to write to
        :param args: additional arguments
        :param kwargs: additional keyword arguments
        :return: the file to write to
        """
        pass



"""
Defines the different dataset classes.

DrugResponseDataset for response values and FeatureDataset for feature values.
They both inherit from the abstract class Dataset.
The DrugResponseDataset class is used
to store drug response values per cell line and drug.
The FeatureDataset class is used to store
feature values per cell line or drug.
The FeatureDataset class can also store meta information
for the feature views. The DrugResponseDataset class
can be split into training, validation and test sets for cross-validation.
The FeatureDataset class can be used to randomize feature vectors.
"""

import copy
import os
from pathlib import Path
from typing import Any, Callable

import networkx as nx
import numpy as np
import pandas as pd
from sklearn.base import TransformerMixin
from sklearn.model_selection import GroupKFold, train_test_split

from ..pipeline_function import pipeline_function
from .utils import permute_features, randomize_graph

np.set_printoptions(threshold=6)


class DrugResponseDataset:
    """Drug response dataset."""

    _response: np.ndarray
    _cell_line_ids: np.ndarray
    _tissues: np.ndarray | None = None
    _drug_ids: np.ndarray
    _predictions: np.ndarray | None = None
    _cv_splits: list[dict[str, "DrugResponseDataset"]] = []
    _name: str

    @classmethod
    def from_csv(
        cls: type["DrugResponseDataset"],
        input_file: str | Path,
        dataset_name: str = "unknown",
        measure: str = "response",
        tissue_column: str | None = None,
    ) -> "DrugResponseDataset":
        """
        Load a dataset from a csv file.

        This function creates a DrugResponseDataset from a provided input file in csv format.
        The following columns are required:
        - response:         the drug response values as floating point values
        - cell_line_ids:    a string identifier for cell lines
        - drug_ids:         a string identifier for drugs
        - predictions:      an optional column containing drug response predictions
        - measure:         the name of the column containing the measure to predict

        :param input_file: Path to the csv file containing the data to be loaded
        :param dataset_name: Optional name to associate the dataset with, default = "unknown"
        :param measure: The name of the column containing the measure to predict, default = "response"
        :param tissue_column: Optional column name of column containing tissue types

        :returns: DrugResponseDataset object containing data from provided csv file.
        """
        data = pd.read_csv(input_file)
        data["drug_id"] = data["drug_id"].astype(str)

        if "predictions" in data.columns:
            predictions = data["predictions"].values
        else:
            predictions = None
        return cls(
            response=data[measure].values,
            cell_line_ids=data["cell_line_id"].values,
            drug_ids=data["drug_id"].values,
            predictions=predictions,
            dataset_name=dataset_name,
            tissues=data[tissue_column].values if tissue_column in data.columns else None,
        )

    @property
    def response(self) -> np.ndarray:
        """
        Returns the response values.

        :returns: numpy array containing response values.
        """
        return self._response

    @property
    def cell_line_ids(self) -> np.ndarray:
        """
        Returns the cell_line_ids.

        :returns: numpy array containing cell_line_ids values.
        """
        return self._cell_line_ids

    @property
    def drug_ids(self) -> np.ndarray:
        """
        Returns the drug_ids.

        :returns: numpy array containing drug_ids values.
        """
        return self._drug_ids

    @property
    def predictions(self) -> np.ndarray | None:
        """
        Returns the predictions if they exist.

        :returns: numpy array containing prediction values or None.
        """
        return self._predictions

    @property
    def tissue(self) -> np.ndarray | None:
        """
        Returns the tissue types if they exist.

        :returns: numpy array containing tissue types or None.
        """
        return self._tissues

    @property
    def cv_splits(self) -> list[dict[str, "DrugResponseDataset"]]:
        """
        Returns the cv_splits.

        :returns: DrugResponseDatasets containing the CV_splits.
        """
        return self._cv_splits

    @property
    def dataset_name(self) -> str:
        """
        Returns the name of this DrugResponseDataset.

        :returns: dataset name.
        """
        return self._name

    @pipeline_function
    def __init__(
        self,
        response: np.ndarray,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        tissues: np.ndarray | None = None,
        predictions: np.ndarray | None = None,
        dataset_name: str = "unnamed",
    ) -> None:
        """
        Initializes the drug response dataset.

        :param response: drug response values per cell line and drug
        :param cell_line_ids: cell line IDs
        :param drug_ids: drug IDs
        :param tissues: Optionally, tissue types of the cell lines for leave-tissue-out cv
        :param predictions: optional. Predicted drug response values per cell line and drug
        :param dataset_name: optional. Name of the dataset, default: "unnamed"
        :raises AssertionError: If response, cell_line_ids, drug_ids, (and the optional predictions) do not all have
            the same length.
        """
        super().__init__()
        if len(response) != len(cell_line_ids):
            raise AssertionError("Response and cell line identifiers have different lengths.")
        if len(response) != len(drug_ids):
            raise AssertionError("Response and drug identifiers have different lengths.")
        if predictions is not None and len(response) != len(predictions):
            raise AssertionError("Response and predictions have different lengths.")
        self._response = response
        self._cell_line_ids = cell_line_ids
        self._drug_ids = drug_ids
        self._predictions = predictions
        self._name = dataset_name
        if tissues is not None:
            self._tissues = np.array(tissues)
        else:
            self._tissues = None

    def __len__(self) -> int:
        """
        Overwrites the default length method.

        :returns: Number of samples in the dataset
        """
        return len(self.response)

    def __str__(self) -> str:
        """
        Overwrite the default str method.

        :return: Text summary of the dataset
        """
        string = (
            f"{self.dataset_name} DrugResponseDataset with {len(self)} entries:\n"
            f"CLs {self.cell_line_ids}\n"
            f"Drugs {self.drug_ids}\n"
            f"Response {self.response}\n"
        )
        if self.predictions is not None:
            string += f"Predictions {self.predictions}\n"
        return string

    def to_dataframe(self) -> pd.DataFrame:
        """
        Convert the dataset into a pandas DataFrame.

        :returns: pandas DataFrame of the dataset with columns 'cell_line_id', 'drug_id', 'response'(, 'predictions')
        """
        data = {
            "cell_line_id": self.cell_line_ids,
            "drug_id": self.drug_ids,
            "response": self.response,
        }
        if self.predictions is not None:
            data["predictions"] = self.predictions
        if self.tissue is not None:
            data["tissue"] = self.tissue

        return pd.DataFrame(data)

    def to_csv(self, path: str | Path):
        """
        Stores the drug response dataset on disk.

        :param path: path to desired storage location
        """
        self.to_dataframe().to_csv(path, index=False)

    @pipeline_function
    def add_rows(self, other: "DrugResponseDataset") -> None:
        """
        Adds rows from another dataset.

        :param other: other dataset
        """
        self._response = np.concatenate([self._response, other.response])
        self._cell_line_ids = np.concatenate([self._cell_line_ids, other.cell_line_ids])
        self._drug_ids = np.concatenate([self._drug_ids, other.drug_ids])

        if self.tissue is not None and other.tissue is not None:
            self._tissues = np.concatenate([self.tissue, other.tissue])

        if self.predictions is not None and other.predictions is not None:
            self._predictions = np.concatenate([self._predictions, other.predictions])

    @pipeline_function
    def remove_nan_responses(self) -> None:
        """Removes rows with NaN values in the response."""
        mask = ~np.isnan(self.response)
        self.mask(mask)

    @pipeline_function
    def shuffle(self, random_state: int = 42) -> None:
        """
        Shuffles the dataset.

        :param random_state: random state
        """
        indices = np.arange(len(self))
        np.random.seed(random_state)
        np.random.shuffle(indices)
        self._response = self.response[indices]
        self._cell_line_ids = self.cell_line_ids[indices]
        self._drug_ids = self.drug_ids[indices]
        if self.predictions is not None:
            self._predictions = self.predictions[indices]
        if self.tissue is not None:
            self._tissues = self.tissue[indices]

    def _remove_drugs(self, drugs_to_remove: str | list[str | int]) -> None:
        """
        Removes drugs from the dataset.

        :param drugs_to_remove: name of drug or list of names of multiple drugs to remove
        """
        if isinstance(drugs_to_remove, str):
            drugs_to_remove = [drugs_to_remove]

        mask = np.array([drug not in drugs_to_remove for drug in self.drug_ids], dtype=bool)
        self.mask(mask)

    def _remove_cell_lines(self, cell_lines_to_remove: str | list[str | int]) -> None:
        """
        Removes cell lines from the dataset.

        :param cell_lines_to_remove: name of cell line or list of names of multiple cell lines to remove
        """
        if isinstance(cell_lines_to_remove, str):
            cell_lines_to_remove = [cell_lines_to_remove]

        mask = np.array([cell_line not in cell_lines_to_remove for cell_line in self.cell_line_ids], dtype=bool)
        self.mask(mask)

    def remove_rows(self, indices: np.ndarray) -> None:
        """
        Removes rows from the dataset.

        :param indices: indices of rows to remove
        """
        mask = np.ones(len(self), dtype=bool)
        mask[indices] = False
        self.mask(mask)

    def reduce_to(self, cell_line_ids: np.ndarray | None = None, drug_ids: np.ndarray | None = None) -> None:
        """
        Removes all rows which contain a cell_line not in cell_line_ids or a drug not in drug_ids.

        :param cell_line_ids: cell line IDs or None to keep all cell lines
        :param drug_ids: drug IDs or None to keep all cell lines
        """
        if drug_ids is not None:
            self._remove_drugs(list(set(self.drug_ids) - set(drug_ids)))

        if cell_line_ids is not None:
            self._remove_cell_lines(list(set(self.cell_line_ids) - set(cell_line_ids)))

    @pipeline_function
    def split_dataset(
        self,
        n_cv_splits: int,
        mode: str,
        split_validation: bool = True,
        split_early_stopping: bool = True,
        validation_ratio: float = 0.1,
        random_state: int = 42,
    ) -> list[dict]:
        """
        Splits the dataset into training, validation and test sets for cross-validation.

        :param n_cv_splits: number of cross-validation splits, e.g., 5
        :param mode: split mode ('LPO', 'LCO', 'LDO')
        :param split_validation: if True, a validation set is generated
        :param split_early_stopping: if True, an early stopping set is generated
        :param validation_ratio: ratio of validation set size to training set size
        :param random_state: random state
        :returns: list of dictionaries containing the cross-validation datasets.
            Each fold is a dictionary with keys 'train', 'validation', 'test', 'validation_es', 'early_stopping'.
        :raises ValueError: if mode is not 'LPO', 'LCO', or 'LDO'
        :raises ValueError: if LTO cross-validation but tissue information not provided
        """
        if mode == "LPO":
            cv_splits = _leave_pair_out_cv(
                n_cv_splits=n_cv_splits,
                response=self.response,
                cell_line_ids=self.cell_line_ids,
                drug_ids=self.drug_ids,
                tissues=self.tissue,
                split_validation=split_validation,
                validation_ratio=validation_ratio,
                random_state=random_state,
                dataset_name=self.dataset_name,
            )

        elif mode in ["LCO", "LTO", "LDO"]:
            if mode == "LTO":
                # Leave-tissue-out cross-validation
                group = "tissue"
                if self.tissue is None:
                    raise ValueError("Tissue information is required for LTO cross-validation.")
            elif mode == "LCO":
                # Leave-cell-line-out cross-validation
                group = "cell_line"
            else:
                # Leave-drug-out cross-validation
                group = "drug"

            cv_splits = _leave_group_out_cv(
                group=group,
                n_cv_splits=n_cv_splits,
                response=self.response,
                cell_line_ids=self.cell_line_ids,
                drug_ids=self.drug_ids,
                tissues=self.tissue,
                split_validation=split_validation,
                validation_ratio=validation_ratio,
                random_state=random_state,
                dataset_name=self.dataset_name,
            )
        else:
            raise ValueError(f"Unknown split mode {mode!r}. Choose from 'LPO', 'LCO', 'LTO', 'LDO'.")

        if split_validation and split_early_stopping:
            for split in cv_splits:
                validation_es, early_stopping = _split_early_stopping_data(split["validation"], test_mode=mode)
                split["validation_es"] = validation_es
                split["early_stopping"] = early_stopping
        self._cv_splits = cv_splits
        return cv_splits

    def save_splits(self, path: str):
        """
        Save cross validation splits to path/cv_split_0_train.csv and path/cv_split_0_test.csv.

        :param path: path to the directory where the cv split files are saved
        :raises AssertionError: if DrugResponseDataset was not split
        """
        if not self.cv_splits:
            raise AssertionError("Trying to save splits, but DrugResponseDataset was not split.")
        os.makedirs(path, exist_ok=True)
        for i, split in enumerate(self.cv_splits):

            for mode in [
                "train",
                "validation",
                "test",
                "validation_es",
                "early_stopping",
            ]:
                if mode in split:
                    split_path = os.path.join(path, f"cv_split_{i}_{mode}.csv")
                    split[mode].to_csv(path=split_path)

    def load_splits(self, path: str) -> None:
        """
        Load cross validation splits from path/cv_split_0_train.csv and path/cv_split_0_test.csv.

        :param path: path to the directory containing the cv split files
        :raises AssertionError: if no cv split files are found in path
        """
        files = os.listdir(path)
        files = [file for file in files if (file.endswith(".csv") and file.startswith("cv_split"))]
        if len(files) == 0:
            raise AssertionError(f"No cv split files found in {path}")

        train_splits = [file for file in files if "train" in file]
        test_splits = [file for file in files if "test" in file]

        validation_es_splits = [file for file in files if "validation_es" in file]
        validation_splits = [file for file in files if "validation" in file and file not in validation_es_splits]
        early_stopping_splits = [file for file in files if "early_stopping" in file]

        for ds in [
            train_splits,
            test_splits,
            validation_splits,
            validation_es_splits,
            early_stopping_splits,
        ]:
            ds.sort()

        optional_splits = {
            "validation": validation_splits,
            "validation_es": validation_es_splits,
            "early_stopping": early_stopping_splits,
        }
        self._cv_splits.clear()  # TODO do we need this?

        for split_train, split_test in zip(train_splits, test_splits, strict=True):
            tr_split = DrugResponseDataset.from_csv(os.path.join(path, split_train), dataset_name=self.dataset_name)
            te_split = DrugResponseDataset.from_csv(os.path.join(path, split_test), dataset_name=self.dataset_name)
            self._cv_splits.append({"train": tr_split, "test": te_split})

        for mode in ["validation", "validation_es", "early_stopping"]:
            if len(optional_splits[mode]) > 0:
                for i, v_split in enumerate(optional_splits[mode]):
                    split = DrugResponseDataset.from_csv(os.path.join(path, v_split), dataset_name=self.dataset_name)
                    self._cv_splits[i][mode] = split

    def copy(self):
        """Returns a copy of the drug response dataset.

        :returns: copy of the dataset
        """
        return DrugResponseDataset(
            response=copy.deepcopy(self.response),
            cell_line_ids=copy.deepcopy(self.cell_line_ids),
            drug_ids=copy.deepcopy(self.drug_ids),
            predictions=copy.deepcopy(self.predictions),
            tissues=copy.deepcopy(self.tissue),
            dataset_name=self.dataset_name,
        )

    def __hash__(self) -> int:
        """Overwrites default hash method.

        :returns: hash value of the dataset
        """
        return hash(
            (
                self.dataset_name,
                tuple(self.cell_line_ids),
                tuple(self.drug_ids),
                tuple(self.response),
                (tuple(self.predictions) if self.predictions is not None else None),
                (tuple(self.tissue) if self.tissue is not None else None),
            )
        )

    def mask(self, mask: np.ndarray) -> None:
        """
        Removes rows from the dataset based on a boolean mask.

        :param mask: boolean mask
        :raises ValueError: if mask is not boolean or integer
        """
        if mask.dtype != bool and not np.issubdtype(mask.dtype, np.integer):
            raise ValueError("Mask must be of boolean or integer dtype.")

        self._response = self.response[mask]
        self._cell_line_ids = self.cell_line_ids[mask]
        self._drug_ids = self.drug_ids[mask]
        if self.predictions is not None:
            self._predictions = self.predictions[mask]
        if self.tissue is not None:
            self._tissues = self.tissue[mask]

    def transform(self, response_transformation: TransformerMixin) -> None:
        """
        Apply transformation to the response data and prediction data of the dataset.

        :param response_transformation: e.g., StandardScaler, MinMaxScaler, RobustScaler
        """
        self._response = response_transformation.transform(self.response.reshape(-1, 1)).squeeze()
        if self.predictions is not None:
            self._predictions = response_transformation.transform(self.predictions.reshape(-1, 1)).squeeze()

    def fit_transform(self, response_transformation: TransformerMixin) -> None:
        """
        Fit and transform the response data and prediction data of the dataset.

        :param response_transformation: e.g., StandardScaler, MinMaxScaler, RobustScaler
        """
        response_transformation.fit(self.response.reshape(-1, 1))
        self.transform(response_transformation)

    def inverse_transform(self, response_transformation: TransformerMixin) -> None:
        """
        Inverse transform the response data and prediction data of the dataset.

        :param response_transformation: e.g., StandardScaler, MinMaxScaler, RobustScaler
        """
        self._response = response_transformation.inverse_transform(self.response.reshape(-1, 1)).squeeze()
        if self.predictions is not None:
            self._predictions = response_transformation.inverse_transform(self.predictions.reshape(-1, 1)).squeeze()


def _split_early_stopping_data(
    validation_dataset: DrugResponseDataset, test_mode: str
) -> tuple[DrugResponseDataset, DrugResponseDataset]:
    """
    Splits the validation dataset into a validation and an early stopping dataset.

    :param validation_dataset: input validation dataset
    :param test_mode: LPO, LCO, LTO, LDO
    :raises ValueError: if test_mode is not one of the expected values
    :returns: the resulting validation and early stopping datasets
    """
    validation_dataset.shuffle(random_state=42)

    # Determine the number of splits b (default 4,
    # but can be less if there are not enough groups)
    if test_mode == "LTO":
        tissues = validation_dataset.tissue
        if tissues is None:
            raise ValueError("Tissue information is required for LTO.")
        n_splits = min(4, len(np.unique(tissues)))
    elif test_mode == "LCO":
        n_splits = min(4, len(np.unique(validation_dataset.cell_line_ids)))
    elif test_mode == "LDO":
        n_splits = min(4, len(np.unique(validation_dataset.drug_ids)))
    else:
        n_splits = 4

    cv_v = validation_dataset.split_dataset(
        n_cv_splits=n_splits,
        mode=test_mode,
        split_validation=False,
        split_early_stopping=False,
        random_state=42,
    )
    # take the first fold of a 4 cv as the split i.e. 3/4 for validation and 1/4 for early stopping
    # when n_groups is less than 4, we splits the validation dataset into 2/3 and 1/3 or 1/2 and 1/2
    validation_dataset = cv_v[0]["train"]
    early_stopping_dataset = cv_v[0]["test"]
    return validation_dataset, early_stopping_dataset


def _leave_pair_out_cv(
    n_cv_splits: int,
    response: np.ndarray,
    cell_line_ids: np.ndarray,
    drug_ids: np.ndarray,
    tissues: np.ndarray | None = None,
    split_validation: bool = True,
    validation_ratio: float = 0.1,
    random_state: int = 42,
    dataset_name: str = "unknown",
) -> list[dict[str, DrugResponseDataset]]:
    """
    Leave pair out cross validation. Splits data into n_cv_splits number of cross validation splits.

    :param n_cv_splits: number of cross validation splits
    :param response: response (e.g. ic50 values)
    :param cell_line_ids: cell line IDs
    :param drug_ids: drug IDs
    :param tissues: tissue types of the cell line, if available
    :param split_validation: whether to split the training set into training and validation set
    :param validation_ratio: ratio of validation set (of the training set)
    :param random_state: random state
    :param dataset_name: name of the dataset
    :returns: list of dicts of the cross validation sets
    :raises AssertionError: if response, cell_line_ids and drug_ids have different lengths
    """
    if not (len(response) == len(cell_line_ids) == len(drug_ids)):
        raise AssertionError("response, cell_line_ids and drug_ids must have the same length")
    indices = np.arange(len(response))
    np.random.seed(random_state)
    shuffled_indices = np.random.permutation(indices)
    response = response[shuffled_indices].copy()
    cell_line_ids = cell_line_ids[shuffled_indices].copy()
    drug_ids = drug_ids[shuffled_indices].copy()
    if tissues is not None:
        tissues = tissues[shuffled_indices].copy()

    # We use GroupKFold to ensure that each pair is only in one fold (prevent data leakage due to
    # experimental replicates).
    # If there are no replicates this is equivalent to KFold.
    groups = [cell + "_" + drug for cell, drug in zip(cell_line_ids, drug_ids, strict=True)]
    kf = GroupKFold(n_splits=n_cv_splits)
    cv_sets = []

    for train_indices, test_indices in kf.split(response, groups=groups):
        if split_validation:
            # split training set into training and validation set
            train_indices, validation_indices = train_test_split(
                train_indices,
                test_size=validation_ratio,
                shuffle=True,
                random_state=random_state,
            )
        cv_fold = {
            "train": DrugResponseDataset(
                cell_line_ids=cell_line_ids[train_indices],
                drug_ids=drug_ids[train_indices],
                response=response[train_indices],
                tissues=tissues[train_indices] if tissues is not None else None,
                dataset_name=dataset_name,
            ),
            "test": DrugResponseDataset(
                cell_line_ids=cell_line_ids[test_indices],
                drug_ids=drug_ids[test_indices],
                response=response[test_indices],
                tissues=tissues[test_indices] if tissues is not None else None,
                dataset_name=dataset_name,
            ),
        }

        if split_validation:
            cv_fold["validation"] = DrugResponseDataset(
                cell_line_ids=cell_line_ids[validation_indices],
                drug_ids=drug_ids[validation_indices],
                response=response[validation_indices],
                tissues=tissues[validation_indices] if tissues is not None else None,
                dataset_name=dataset_name,
            )

        cv_sets.append(cv_fold)
    return cv_sets


def _leave_group_out_cv(
    group: str,
    n_cv_splits: int,
    response: np.ndarray,
    cell_line_ids: np.ndarray,
    drug_ids: np.ndarray,
    tissues: np.ndarray | None = None,
    split_validation: bool = True,
    validation_ratio: float = 0.1,
    random_state: int = 42,
    dataset_name: str = "unknown",
):
    """
    Leave group out cross validation: Splits data into n_cv_splits number of cross validation splits.

    :param group: group to leave out (cell_line or drug)
    :param n_cv_splits: number of cross validation splits
    :param response: response (e.g. ic50 values)
    :param cell_line_ids: cell line IDs
    :param drug_ids: drug IDs
    :param tissues: tissue types of the cell line, if available (required for LTO)
    :param split_validation: whether to split the training set into training and validation set
    :param validation_ratio: ratio of validation set (of the training set)
    :param random_state: random state
    :param dataset_name: name of the dataset
    :returns: list of dicts of the cross validation sets
    :raises AssertionError: if group is not 'cell_line' or 'drug' or 'tissue'
    :raises AssertionError: Tissue information is required for LTO cross-validation
    """
    if group not in {"cell_line", "drug", "tissue"}:
        raise AssertionError(f"group must be 'cell_line' or 'drug', but is {group}")

    if group == "cell_line":
        group_ids = cell_line_ids
    elif group == "drug":
        group_ids = drug_ids
    elif group == "tissue":
        if tissues is None:
            raise AssertionError("Tissue information is required for LTO cross-validation.")
        group_ids = tissues
    else:
        raise AssertionError(f"Unknown group {group}")

    # shuffle, since GroupKFold does not implement this
    indices = np.arange(len(response))
    np.random.seed(random_state)
    shuffled_indices = np.random.permutation(indices)
    response = response[shuffled_indices].copy()
    cell_line_ids = cell_line_ids[shuffled_indices].copy()
    drug_ids = drug_ids[shuffled_indices].copy()
    tissues = tissues[shuffled_indices].copy() if tissues is not None else None
    group_ids = group_ids[shuffled_indices].copy()
    gkf = GroupKFold(n_splits=n_cv_splits)
    cv_sets = []

    for train_indices, test_indices in gkf.split(response, groups=group_ids):
        cv_fold = {
            "train": DrugResponseDataset(
                cell_line_ids=cell_line_ids[train_indices],
                drug_ids=drug_ids[train_indices],
                response=response[train_indices],
                tissues=tissues[train_indices] if tissues is not None else None,
                dataset_name=dataset_name,
            ),
            "test": DrugResponseDataset(
                cell_line_ids=cell_line_ids[test_indices],
                drug_ids=drug_ids[test_indices],
                response=response[test_indices],
                tissues=tissues[test_indices] if tissues is not None else None,
                dataset_name=dataset_name,
            ),
        }
        if split_validation:
            # split training set into training and validation set.
            # The validation set also does
            # contain unqiue cell lines/drugs
            unique_train_groups = np.unique(group_ids[train_indices])
            train_groups, validation_groups = train_test_split(
                unique_train_groups,
                test_size=validation_ratio,
                shuffle=True,
                random_state=random_state,
            )
            train_indices = np.where(np.isin(group_ids, train_groups))[0]
            validation_indices = np.where(np.isin(group_ids, validation_groups))[0]
            cv_fold["train"] = DrugResponseDataset(
                cell_line_ids=cell_line_ids[train_indices],
                drug_ids=drug_ids[train_indices],
                response=response[train_indices],
                tissues=tissues[train_indices] if tissues is not None else None,
                dataset_name=dataset_name,
            )
            cv_fold["validation"] = DrugResponseDataset(
                cell_line_ids=cell_line_ids[validation_indices],
                drug_ids=drug_ids[validation_indices],
                response=response[validation_indices],
                tissues=tissues[validation_indices] if tissues is not None else None,
                dataset_name=dataset_name,
            )

        cv_sets.append(cv_fold)
    return cv_sets


class FeatureDataset:
    """Class for feature datasets."""

    _meta_info: dict[str, Any] = {}
    _features: dict[str, dict[str, Any]] = {}

    @classmethod
    def from_csv(
        cls: type["FeatureDataset"],
        path_to_csv: str | Path,
        id_column: str,
        view_name: str,
        drop_columns: list[str] | None = None,
    ):
        """Load a one-view feature dataset from a csv file.

        Load a feature dataset from a csv file. The rows of the csv file represent the instances (cell lines or drugs),
        the columns represent the features. A column named id_column contains the identifiers of the instances.
        All unrelated columns (e.g. other id columns) should be provided as drop_columns,
        that will be removed from the dataset.

        :param path_to_csv: path to the csv file containing the data to be loaded
        :param view_name: name of the view (e.g. gene_expression)
        :param id_column: name of the column containing the identifiers
        :param drop_columns: list of columns to drop (e.g. other identifier columns)
        :returns: FeatureDataset object containing data from provided csv file.
        """
        data = pd.read_csv(path_to_csv)
        ids = data[id_column].values
        data_features = data.drop(columns=(drop_columns or []))
        data_features = data_features.set_index(id_column)
        # remove duplicate feature rows (rows with the same index)
        data_features = data_features[~data_features.index.duplicated(keep="first")]
        features = {}

        for identifier in ids:
            features_for_instance = data_features.loc[identifier].values
            features[identifier] = {view_name: features_for_instance}

        return cls(features=features)

    def to_csv(self, path: str | Path, id_column: str, view_name: str):
        """
        Save the feature dataset to a CSV file.

        :param path: Path to the CSV file.
        :param id_column: Name of the column containing the identifiers.
        :param view_name: Name of the view (e.g., gene_expression).

        :raises ValueError: If the view is not found for an identifier.
        """
        data = []
        for identifier, feature_dict in self.features.items():
            # Get the feature vector for the specified view
            if view_name in feature_dict:
                row = {id_column: identifier}
                row.update({f"feature_{i}": value for i, value in enumerate(feature_dict[view_name])})
                data.append(row)
            else:
                raise ValueError(f"View {view_name!r} not found for identifier {identifier!r}.")

        # Convert to DataFrame and save to CSV
        df = pd.DataFrame(data)
        df.to_csv(path, index=False)

    @property
    def meta_info(self) -> dict[str, Any]:
        """
        Returns the meta information.

        :returns: Meta information of this FeatureDataset
        """
        return self._meta_info

    @property
    def features(self) -> dict[str, dict[str, Any]]:
        """
        Returns the features.

        :returns: features of this FeatureDataset
        """
        return self._features

    @property
    def identifiers(self) -> np.ndarray:
        """
        Returns the identifiers of the features.

        :returns: feature identifiers of this FeatureDataset
        """
        return np.array(list(self.features.keys()))

    @property
    def view_names(self) -> list[str]:
        """
        Returns the view_names.

        :returns: view_names of this FeatureDataset
        """
        return list(self.features[list(self.features.keys())[0]].keys())  # TODO whut?!

    def __init__(
        self,
        features: dict[str, dict[str, Any]],
        meta_info: dict[str, Any] | None = None,
    ):
        """
        Initializes the feature dataset.

        :param features: dictionary of features,
            key: drug ID/cell line ID, value: Dict of feature views,
            key: feature name, value: feature vector
        :param meta_info: additional information for the views, e.g. gene names for gene expression
        :raises AssertionError: if meta_info keys are not in view names
        """
        super().__init__()
        self._features = features
        if meta_info is not None:
            # assert that str of meta Dict[str, Any] is in view_names
            if not all(meta_key in self.view_names for meta_key in meta_info.keys()):
                raise AssertionError(f"Meta keys {meta_info.keys()} not in view names {self.view_names}")
            self._meta_info = meta_info

    def randomize_features(self, views_to_randomize: str | list[str], randomization_type: str) -> None:
        """
        Randomizes the feature vectors.

        Permutation permutes the feature vectors.
        Invariant means that the randomization is done in a way that a key characteristic of the feature is
        preserved. In case of matrices, this is the mean and standard deviation of the feature view for this
        instance, for networks it is the degree distribution.

        :param views_to_randomize: name of feature view or list of names of multiple feature views
            to randomize. The other views are not randomized.
        :param randomization_type: randomization type ('permutation', 'invariant').
        :raises AssertionError: if randomization_type is not 'permutation' or 'invariant'
        :raises ValueError: if no invariant randomization is available for the feature view type
        """
        if randomization_type not in ["permutation", "invariant"]:
            raise AssertionError(
                f"Unknown randomization type {randomization_type!r}. Choose from 'permutation', 'invariant'."
            )

        if isinstance(views_to_randomize, str):
            views_to_randomize = [views_to_randomize]

        if randomization_type == "permutation":
            # Permute the specified views for each entity (= cell line or drug)
            # E.g. each cell line gets the feature vector/graph/image...
            # of another cell line.
            # Drawn without replacement.
            self._features = permute_features(
                features=self.features,
                views_to_permute=views_to_randomize,
                identifiers=self.identifiers,
                all_views=self.view_names,
            )

        elif randomization_type == "invariant":
            # Invariant randomization:
            # Randomize the specified views for each entity in a way that
            # a key characteristic of the feature is preserved.
            # For vectors this is the mean and standard deviation the feature view,
            # for networks the degree distribution.
            for view in views_to_randomize:
                for identifier in self.identifiers:
                    if isinstance(self.features[identifier][view], np.ndarray):
                        new_features = np.random.normal(
                            self.features[identifier][view].mean(),
                            self.features[identifier][view].std(),
                            self.features[identifier][view].shape,
                        )
                    elif isinstance(self.features[identifier][view], nx.classes.graph.Graph):
                        new_features = randomize_graph(self.features[identifier][view])

                    else:
                        raise ValueError(
                            f"No invariant randomization available for feature view "
                            f"type {type(self.features[identifier][view])!r}."
                        )
                    self.features[identifier][view] = new_features

    def get_feature_matrix(self, view: str, identifiers: np.ndarray) -> np.ndarray:
        """
        Returns the feature matrix for the given view.

        The feature view must be a vector or matrix.
        :param view: view name
        :param identifiers: list of identifiers (cell lines oder drugs)
        :returns: feature matrix
        :raises AssertionError: if no identifiers are given
        :raises AssertionError: if view is not in the FeatureDataset
        :raises AssertionError: if identifiers are not in the FeatureDataset
        :raises AssertionError: if feature vectors of view have different lengths
        :raises AssertionError: if view is not a numpy array, i.e. not a vector or matrix
        """
        if len(identifiers) == 0:
            raise AssertionError("get_feature_matrix: No identifiers given.")

        if view not in self.view_names:
            raise AssertionError(f"View {view!r} not in in the FeatureDataset.")
        missing_identifiers = {id_ for id_ in identifiers if id_ not in self.identifiers}
        if missing_identifiers:
            raise AssertionError(
                f"{len(missing_identifiers)} of {len(np.unique(identifiers))} ids are not in the "
                f"FeatureDataset. Missing ids: {missing_identifiers}"
            )

        if not all(len(self.features[id_][view]) == len(self.features[identifiers[0]][view]) for id_ in identifiers):
            raise AssertionError(f"Feature vectors of view {view} have different lengths.")

        if not all(isinstance(self.features[id_][view], np.ndarray) for id_ in identifiers):
            raise AssertionError(f"get_feature_matrix only works for vectors or matrices. {view} is not a numpy array.")
        out = np.array([self.features[id_][view] for id_ in identifiers])
        return out

    def copy(self):
        """Returns a copy of the feature dataset.

        :returns: copy of the dataset
        """
        return FeatureDataset(features=copy.deepcopy(self.features))

    def add_features(self, other: "FeatureDataset") -> None:
        """
        Adds features views from another dataset. Inner join (only common identifiers are kept).

        :param other: other dataset
        :raises AssertionError: if feature views overlap
        """
        if len(set(self.view_names) & set(other.view_names)) != 0:
            raise AssertionError(
                "Trying to add features but feature views overlap. FeatureDatasets should be distinct."
            )
        if other.meta_info:
            self.add_meta_info(other)

        common_identifiers = set(self.identifiers).intersection(other.identifiers)
        new_features = {}
        for id_ in common_identifiers:
            id_ = str(id_)
            new_features[id_] = {view: self.features[id_][view] for view in self.view_names}
            for view in other.view_names:
                new_features[id_][view] = other.features[id_][view]

        self._features = new_features

    def add_meta_info(self, other: "FeatureDataset") -> None:
        """
        Adds meta information to the feature dataset.

        :param other: other dataset
        """
        other_meta = other.meta_info
        if self.meta_info is None:
            self.meta_info = other_meta
        else:
            if other_meta is not None:
                self.meta_info.update(other_meta)

    def transform_features(self, ids: np.ndarray, transformer: TransformerMixin, view: str):
        """
        Applies a transformation like standard scaling to features.

        :param ids: The IDs to transform
        :param transformer: fitted sklearn transformer
        :param view: the view to transform
        :raises AssertionError: if view is not in the FeatureDataset
        :raises AssertionError: if a cell line is missing
        :raises AssertionError: if IDs are not unique
        """
        if view not in self.view_names:
            raise AssertionError(f"Transform view {view!r} not in in the FeatureDataset.")
        if not all([clid in self.features for clid in ids]):
            raise AssertionError("Trying to transform, but a cell line is missing.")

        if len(np.unique(ids)) != len(ids):
            raise AssertionError("IDs should be unique.")

        for identifier in ids:
            feature_vector = self.features[identifier][view]
            scaled_feature_vector = transformer.transform([feature_vector])[0]
            self.features[identifier][view] = scaled_feature_vector

    def fit_transform_features(self, train_ids: np.ndarray, transformer: TransformerMixin, view: str):
        """
        Fits and applies a transformation. Fitting is done only on the train_ids.

        :param train_ids: The IDs corresponding to the training dataset.
        :param transformer: sklearn transformer
        :param view: the view to transform
        :returns: The modified FeatureDataset with transformed gene expression features.
        :raises AssertionError: if view is not in the FeatureDataset
        :raises AssertionError: if train IDs are not unique
        """
        if view not in self.view_names:
            raise AssertionError(f"Transform view {view!r} not in in the FeatureDataset.")

        if len(np.unique(train_ids)) != len(train_ids):
            raise AssertionError("Train IDs should be unique.")

        # Collect all features of the view for fitting the scaler
        train_features = np.vstack([self.features[identifier][view] for identifier in train_ids])
        transformer.fit(train_features)

        # Apply transformation and scaling to each feature vector
        for identifier in self.features:
            feature_vector = self.features[identifier][view]
            scaled_gene_expression = transformer.transform([feature_vector])[0]
            self.features[identifier][view] = scaled_gene_expression
        return transformer

    def apply(self, function: Callable, view: str):
        """Applies a function to the features of a view.

        :param function: function to apply
        :param view: view to apply the function to
        """
        for identifier in self.features:
            self.features[identifier][view] = function(self.features[identifier][view])



"""Module containing all drug response prediction models."""

__all__ = [
    "MULTI_DRUG_MODEL_FACTORY",
    "SINGLE_DRUG_MODEL_FACTORY",
    "MODEL_FACTORY",
    "NaivePredictor",
    "NaiveDrugMeanPredictor",
    "NaiveCellLineMeanPredictor",
    "NaiveTissueMeanPredictor",
    "NaiveMeanEffectsPredictor",
    "ElasticNetModel",
    "RandomForest",
    "SVMRegressor",
    "SimpleNeuralNetwork",
    "MultiOmicsNeuralNetwork",
    "MultiOmicsRandomForest",
    "SingleDrugRandomForest",
    "SingleDrugElasticNet",
    "SingleDrugProteomicsElasticNet",
    "SRMF",
    "GradientBoosting",
    "MOLIR",
    "SuperFELTR",
    "DIPKModel",
    "ProteomicsRandomForest",
]

from .baselines.multi_omics_random_forest import MultiOmicsRandomForest
from .baselines.naive_pred import (
    NaiveCellLineMeanPredictor,
    NaiveDrugMeanPredictor,
    NaiveMeanEffectsPredictor,
    NaivePredictor,
    NaiveTissueMeanPredictor,
)
from .baselines.singledrug_elastic_net import SingleDrugElasticNet, SingleDrugProteomicsElasticNet
from .baselines.singledrug_random_forest import SingleDrugRandomForest
from .baselines.sklearn_models import (
    ElasticNetModel,
    GradientBoosting,
    ProteomicsRandomForest,
    RandomForest,
    SVMRegressor,
)
from .DIPK.dipk import DIPKModel
from .drp_model import DRPModel
from .MOLIR.molir import MOLIR
from .SimpleNeuralNetwork.multiomics_neural_network import MultiOmicsNeuralNetwork
from .SimpleNeuralNetwork.simple_neural_network import SimpleNeuralNetwork
from .SRMF.srmf import SRMF
from .SuperFELTR.superfeltr import SuperFELTR

# SINGLE_DRUG_MODEL_FACTORY is used in the pipeline!
SINGLE_DRUG_MODEL_FACTORY: dict[str, type[DRPModel]] = {
    "SingleDrugRandomForest": SingleDrugRandomForest,
    "MOLIR": MOLIR,
    "SuperFELTR": SuperFELTR,
    "SingleDrugElasticNet": SingleDrugElasticNet,
    "SingleDrugProteomicsElasticNet": SingleDrugProteomicsElasticNet,
}

# MULTI_DRUG_MODEL_FACTORY is used in the pipeline!
MULTI_DRUG_MODEL_FACTORY: dict[str, type[DRPModel]] = {
    "NaivePredictor": NaivePredictor,
    "NaiveDrugMeanPredictor": NaiveDrugMeanPredictor,
    "NaiveCellLineMeanPredictor": NaiveCellLineMeanPredictor,
    "NaiveMeanEffectsPredictor": NaiveMeanEffectsPredictor,
    "NaiveTissueMeanPredictor": NaiveTissueMeanPredictor,
    "ElasticNet": ElasticNetModel,
    "RandomForest": RandomForest,
    "SVR": SVMRegressor,
    "SimpleNeuralNetwork": SimpleNeuralNetwork,
    "MultiOmicsNeuralNetwork": MultiOmicsNeuralNetwork,
    "MultiOmicsRandomForest": MultiOmicsRandomForest,
    "GradientBoosting": GradientBoosting,
    "SRMF": SRMF,
    "DIPK": DIPKModel,
    "ProteomicsRandomForest": ProteomicsRandomForest,
}

# MODEL_FACTORY is used in the pipeline!
MODEL_FACTORY = MULTI_DRUG_MODEL_FACTORY.copy()
MODEL_FACTORY.update(SINGLE_DRUG_MODEL_FACTORY)
